#!/usr/bin/env python

"""ALLFIT_NUMBA.PY - Fit PSF to all stars in an image

"""

__authors__ = 'David Nidever <dnidever@montana.edu?'
__version__ = '20241013'  # yyyymmdd


import os
import numpy as np
import time
from numba import njit,types,from_dtype
from numba.experimental import jitclass
from . import utils_numba as utils, groupfit_numba as gfit, models_numba as mnb
from .clock_numba import clock

@njit(cache=True)
def skyval(array,sigma):
    """  Estimate sky value from sky pixels."""
    wt = 1/sigma**2
    med = np.median(array)
    sig = utils.mad(array)
    # reweight outlier pixels using Stetson's method
    resid = array-med
    wt2 = wt/(1+np.abs(resid)**2/np.median(sigma))
    xmn = np.sum(wt2*array)/np.sum(wt2)
    return xmn

    
kv_ty = (types.int64, types.unicode_type)
spec = [
    ('psftype', types.int32),
    ('psfparams', types.float64[:]),
    ('verbose', types.boolean),
    ('nofreeze', types.boolean),
    ('psflookup', types.float64[:,:,:]),
    ('psforder', types.int32),
    #('fwhm', types.float64),
    ('image', types.float64[:,:]),
    ('error', types.float64[:,:]),
    ('mask', types.boolean[:,:]),
    ('skyim', types.float64[:]),
    ('modelim', types.float64[:]),
    ('resid', types.float64[:]),
    ('xx', types.int64[:]),
    ('yy', types.int64[:]),
    ('tab', types.float64[:,:]),
    ('initpars', types.float64[:,:]),
    ('pars', types.float64[:]),
    ('perror', types.float64[:]),
    ('nstars', types.int32),
    ('niter', types.int32),    
    ('npsfpix', types.int32),    
    ('nx', types.int32),
    ('ny', types.int32),
    ('imshape', types.int64[:]),
    ('npix', types.int32),    
    ('fitradius', types.float64),
    ('skyradius', types.float64),
    ('nfitpix', types.int32),
    ('radius', types.int32),
    ('_starsky', types.float64[:]),
    ('starniter', types.int32[:]),
    ('njaciter', types.int32),
    ('freezestars', types.boolean[:]),
    ('starchisq', types.float64[:]),
    ('starrms', types.float64[:]),
    ('star_ravelindex', types.int64[:,:]),
    ('star_ndata', types.int32[:]),
    ('starfit_ravelindex', types.int64[:,:]),
    ('starfit_ndata', types.int32[:]),
    ('starsky_ravelindex', types.int64[:,:]),
    ('starsky_ndata', types.int32[:]),
    ('ntotpix', types.int32),
    ('ravelindex', types.int64[:]),
]

@jitclass(spec)
#@jitclass(spec, cache=True)
class AllFitter(object):

    def __init__(self,psftype,psfparams,npix,psflookup,image,error,mask,
                 tab,fitradius,verbose=False,nofreeze=False):
        # Save the input values
        self.psftype = psftype
        self.psfparams = psfparams
        self.psflookup = psflookup
        _,_,npsforder = psflookup.shape
        self.psforder = 0
        if npsforder==4:
            self.psforder = 1
        if psflookup.ndim != 3:
            raise Exception('psflookup must have 3 dimensions')
        psforder = psflookup.shape[2]
        if psforder>1:
            self.psforder = 1
        self.verbose = verbose
        self.nofreeze = nofreeze
        ny,nx = image.shape                             # save image dimensions, python images are (Y,X)
        self.nx = nx
        self.ny = ny
        self.imshape = np.array([ny,nx])
        self.image = image.copy().astype(np.float64)
        self.error = error.astype(np.float64)
        self.mask = mask.astype(np.bool_)
        xx,yy = utils.meshgrid(np.arange(nx),np.arange(ny))
        self.xx = xx.flatten()
        self.yy = yy.flatten()
        # Order stars by flux, brightest first
        si = np.argsort(tab[:,1])[::-1]  # largest amp first
        self.tab = tab[si]                              # ID, amp, xcen, ycen
        self.initpars = self.tab[:,1:4]                 # amp, xcen, ycen
        self.nstars = len(self.tab)                     # number of stars
        self.niter = 1                                  # number of iterations in the solver
        self.npsfpix = npix                             # shape of PSF
        self.npix = npix
        self.fitradius = fitradius
        self.nfitpix = int(np.ceil(fitradius))  # +/- nfitpix
        self.skyradius = npix//2 + 10
        # Star amps
        # if 'amp' in cat.colnames:
        #     staramp = cat['amp'].copy()
        # else:
        #     # estimate amp from flux and fwhm
        #     # area under 2D Gaussian is 2*pi*A*sigx*sigy
        #     if 'fwhm' in cat.columns:
        #         amp = cat['flux']/(2*np.pi*(cat['fwhm']/2.35)**2)
        #     else:
        #         amp = cat['flux']/(2*np.pi*(psf.fwhm()/2.35)**2)                
        #     staramp = np.maximum(amp,0)   # make sure it's positive
        # Initialize the parameter array
        pars = np.zeros(self.nstars*3,float) # amp, xcen, ycen
        pars[0::3] = self.initpars[:,0]
        pars[1::3] = self.initpars[:,1]
        pars[2::3] = self.initpars[:,2]
        self.pars = pars
        self.perror = np.zeros(len(pars),np.float64)
        # Sky and Niter arrays for the stars
        self._starsky = np.zeros(self.nstars,float)
        self.starniter = np.zeros(self.nstars,np.int32)
        self.starchisq = np.zeros(self.nstars,np.float64)
        self.starrms = np.zeros(self.nstars,np.float64)
        self.njaciter = 0  # initialize njaciter
        # Initialize the freezestars array
        self.freezestars = np.zeros(self.nstars,np.bool_)
        
        # Get information for all the stars
        xcen = self.pars[1::3]
        ycen = self.pars[2::3]
        hpsfnpix = self.npsfpix//2
        out = self.collatestars(self.imshape,xcen,ycen,hpsfnpix,fitradius,self.skyradius)
        fravelindex,fsndata,ravelindex,sndata,skyravelindex,skyndata = out
        self.star_ravelindex = fravelindex
        self.star_ndata = fsndata
        # Fitting arrays
        self.starfit_ravelindex = ravelindex
        self.starfit_ndata = sndata
        # Sky arrays
        self.starsky_ravelindex = skyravelindex
        self.starsky_ndata = skyndata
        
        # Put indices of all the unique fitted pixels into one array
        ntotpix = np.sum(self.starfit_ndata)
        allravelindex = np.zeros(ntotpix,np.int64)
        count = 0
        for i in range(self.nstars):
            n1 = self.starfit_ndata[i]
            ravelind1 = self.starfit_ravelindex[i,:n1]
            allravelindex[count:count+n1] = ravelind1
            count += n1
        allravelindex = np.unique(allravelindex)
        self.ravelindex = allravelindex
        self.ntotpix = len(allravelindex)
        
        # Create initial smooth sky image
        self.skyim = utils.sky(self.image).flatten()
        # Subtract the initial models from the residual array
        self.resid = image.copy().astype(np.float64).flatten()   # flatten makes it easier to modify
        self.resid[:] -= self.skyim   # subtract smooth sky
        self.modelim = np.zeros(self.imshape[0]*self.imshape[1],np.float64)
        for i in range(self.nstars):
            pars1 = self.starpars[i]
            n1 = self.star_ndata[i]
            ravelind1 = self.starravelindex(i)
            xind1 = self.xx[ravelind1]
            yind1 = self.yy[ravelind1]
            m = self.psf(xind1,yind1,pars1)
            self.resid[ravelind1] -= m
            self.modelim[ravelind1] += m
            

    def getstar(self,imshape,xcen,ycen,hpsfnpix,fitradius,skyradius):
        """ Return a star's full footprint and fitted pixels data."""
        # always return the same size
        # a distance of 6.2 pixels spans 6 full pixels but you could have
        # 0.1 left on one side and 0.1 left on the other side
        # that's why we have to add 2 pixels
        maxpix = int(3.14*skyradius**2)
        nskypix = int(2*skyradius+1)
        skybbox = utils.starbbox((xcen,ycen),imshape,int(skyradius))
        nsx = skybbox[1]-skybbox[0]
        nsy = skybbox[3]-skybbox[2]
        skyravelindex = np.zeros(maxpix,np.int64)-1
        nfpix = hpsfnpix*2+1
        #fbbox = utils.starbbox((xcen,ycen),imshape,hpsfnpix)
        # extra buffer is ALWAYS at the end of each dimension    
        fravelindex = np.zeros(nfpix*nfpix,np.int64)-1
        # Fitting pixels
        npix = int(np.floor(2*fitradius))+2
        bbox = utils.starbbox((xcen,ycen),imshape,fitradius)
        nx = bbox[1]-bbox[0]
        ny = bbox[3]-bbox[2]
        ravelindex = np.zeros(npix*npix,np.int64)-1
        fcount = 0
        count = 0
        skycount = 0
        for j in range(nskypix):
            y = j + skybbox[2]
            for i in range(nskypix):
                x = i + skybbox[0]
                if self.mask[y,x]:   # exclude bad pixels
                    continue
                r = np.sqrt((x-xcen)**2 + (y-ycen)**2)
                if x>=skybbox[0] and x<=skybbox[1]-1 and y>=skybbox[2] and y<=skybbox[3]-1:
                    #if x>=fbbox[0] and x<=fbbox[1]-1 and y>=fbbox[2] and y<=fbbox[3]-1:
                    if r <= 1.0*skyradius and r >= 0.7*skyradius:
                        skymulti_index = (np.array([y]),np.array([x]))
                        skyravelindex[skycount] = utils.ravel_multi_index(skymulti_index,imshape)[0]
                        skycount += 1
                    if r <= 1.0*hpsfnpix:
                        fmulti_index = (np.array([y]),np.array([x]))
                        fravelindex[fcount] = utils.ravel_multi_index(fmulti_index,imshape)[0]
                        fcount += 1
                    if r <= fitradius:
                        multi_index = (np.array([y]),np.array([x]))
                        ravelindex[count] = utils.ravel_multi_index(multi_index,imshape)[0]
                        count += 1
        return (fravelindex,fcount,ravelindex,count,skyravelindex,skycount)

    def collatestars(self,imshape,starx,stary,hpsfnpix,fitradius,skyradius):
        """ Get full footprint and fitted pixels data for all stars."""
        nstars = len(starx)
        nfpix = 2*hpsfnpix+1
        npix = int(np.floor(2*fitradius))+2
        # Full footprint arrays
        fravelindex = np.zeros((nstars,nfpix*nfpix),np.int64)
        fndata = np.zeros(nstars,np.int32)
        # Fitting pixel arrays
        ravelindex = np.zeros((nstars,npix*npix),np.int64)
        ndata = np.zeros(nstars,np.int32)
        skyravelindex = np.zeros((nstars,int(3.14*skyradius**2)),np.int64)
        skyndata = np.zeros(nstars,np.int32)
        for i in range(nstars):
            out = self.getstar(imshape,starx[i],stary[i],hpsfnpix,fitradius,skyradius)
            # full footprint information
            fravelindex1,fn1,ravelindex1,n1,skyravelindex1,skyn1 = out
            fravelindex[i,:] = fravelindex1
            fndata[i] = fn1
            # fitting pixel information
            ravelindex[i,:] = ravelindex1
            ndata[i] = n1
            # sky pixel information
            skyravelindex[i,:] = skyravelindex1
            skyndata[i] = skyn1
        # Trim arrays
        maxfn = np.max(fndata)
        fravelindex = fravelindex[:,:maxfn]
        maxn = np.max(ndata)
        ravelindex = ravelindex[:,:maxn]
        maxskyn = np.max(skyndata)
        skyravelindex = skyravelindex[:,:maxskyn]
    
        return (fravelindex,fndata,ravelindex,ndata,skyravelindex,skyndata)
    
    @property
    def starpars(self):
        """ Return the [amp,xcen,ycen] parameters in [Nstars,3] array.
            You can GET a star's parameters like this:
            pars = self.starpars[4]
            You can also SET a star's parameters a similar way:
            self.starpars[4] = pars
        """
        return self.pars[0:3*self.nstars].copy().reshape(self.nstars,3)
    
    @property
    def staramp(self):
        """ Return the best-fit amps for all stars."""
        return self.pars[0::3]

    @staramp.setter
    def staramp(self,val):
        """ Set staramp values."""
        self.pars[0::3] = val
    
    @property
    def starxcen(self):
        """ Return the best-fit X centers for all stars."""        
        return self.pars[1::3]

    @starxcen.setter
    def starxcen(self,val):
        """ Set starxcen values."""
        self.pars[1::3] = val
    
    @property
    def starycen(self):
        """ Return the best-fit Y centers for all stars."""        
        return self.pars[2::3]    

    @starycen.setter
    def starycen(self,val):
        """ Set starycen values."""
        self.pars[2::3] = val
    
    def sky(self):
        """ (Re)calculate the smooth sky."""
        # Remove the current best-fit model
        resid = self.image-self.modelim.copy().reshape(self.imshape[0],self.imshape[1])    # remove model
        prevsky = self.skyim.copy()
        self.skyim[:] = utils.sky(resid).flatten()
        # Update resid
        self.resid[:] += prevsky
        self.resid[:] -= self.skyim
        
    def starsky(self,i):
        """ Calculate the local sky for a single star."""
        n = self.starsky_ndata[i]
        skyravelindex = self.starsky_ravelindex[i,:n]
        resid = self.resid[skyravelindex]
        err = self.error.ravel()[skyravelindex]
        sky = skyval(resid,err)
        self._starsky[i] = sky
        return sky

    def starnpix(self,i):
        """ Return number of pixels in a star's full circular footprint."""
        return self.star_ndata[i]

    def starravelindex(self,i):
        """ Return ravel index (into full image) of the star's full footprint """
        n = self.star_ndata[i]
        return self.star_ravelindex[i,:n]

    def starfitnpix(self,i):
        """ Return number of fitted pixels for a star."""
        return self.starfit_ndata[i]

    def starfitravelindex(self,i):
        """ Return ravel index (into full image) of the star's fitted pixels."""
        n = self.starfit_ndata[i]
        return self.starfit_ravelindex[i,:n]

    def starfitchisq(self,i):
        """ Return chisq of current best-fit for one star."""
        pars1,xind1,yind1,ravelind1 = self.starfitdata(i)
        n1 = self.starfitnpix(i)
        flux1 = self.image.ravel()[ravelind1].copy()
        err1 = self.error.ravel()[ravelind1].copy()
        model1 = self.psf(xind1,yind1,pars1)
        sky1 = self._starsky[i]
        chisq1 = np.sum((flux1-sky1-model1)**2/err1**2)/n1
        return chisq1
    
    def starfitrms(self,i):
        """ Return rms of current best-fit for one star."""
        pars1,xind1,yind1,ravelind1 = self.starfitdata(i)
        n1 = self.starfitnpix(i)
        flux1 = self.image.ravel()[ravelind1].copy()
        err1 = self.error.ravel()[ravelind1].copy()
        model1 = self.psf(xind1,yind1,pars1)
        sky1 = self._starsky[i]
        # chi value, RMS of the residuals as a fraction of the amp
        rms1 = np.sqrt(np.mean(((flux1-sky1-model1)/pars1[0])**2))
        return rms1
    
    def stardata(self,i):
        """ Return a star's full footprint information."""
        pars = self.pars[i*3:(i+1)*3]
        n = self.star_ndata[i]
        ravelindex = self.star_ravelindex[i,:n]        
        xind = self.xx[ravelindex]
        yind = self.yy[ravelindex]
        return pars,xind,yind,ravelindex
    
    def starfitdata(self,i):
        """ Return a star's fitting pixel information."""
        pars = self.pars[i*3:(i+1)*3]
        n = self.starfit_ndata[i]
        ravelindex = self.starfit_ravelindex[i,:n]
        xind = self.xx[ravelindex]
        yind = self.yy[ravelindex]
        return pars,xind,yind,ravelindex

    @property
    def nfreezestars(self):
        """ Return the number of frozen stars."""
        return np.sum(self.freezestars)

    @property
    def freestars(self):
        """ Return the free stars."""
        return ~self.freezestars
    
    @property
    def nfreestars(self):
        """ Return the number of free stars."""
        return np.sum(self.freestars) 

    def unfreeze(self):
        """ Unfreeze all parameters and stars."""
        self.freezestars = np.zeros(self.nstars,np.bool_)
        ##self.resflat = self.imflat.copy()


    def mkbounds(self,pars,imshape,xoff=10):
        """ Make bounds for a set of input parameters."""
        # is [amp1,xcen1,ycen1,amp2,xcen2,ycen2, ...]

        npars = len(pars)
        ny,nx = imshape
        
        # Make bounds
        lbounds = np.zeros(npars,float)
        ubounds = np.zeros(npars,float)
        lbounds[0:-1:3] = 0
        lbounds[1::3] = np.maximum(pars[1::3]-xoff,0)
        lbounds[2::3] = np.maximum(pars[2::3]-xoff,0)
        lbounds[-1] = -np.inf
        ubounds[0:-1:3] = np.inf
        ubounds[1::3] = np.minimum(pars[1::3]+xoff,nx-1)
        ubounds[2::3] = np.minimum(pars[2::3]+xoff,ny-1)
        ubounds[-1] = np.inf
        
        bounds = (lbounds,ubounds)
                 
        return bounds

    def checkbounds(self,pars,bounds):
        """ Check the parameters against the bounds."""
        # 0 means it's fine
        # 1 means it's beyond the lower bound
        # 2 means it's beyond the upper bound
        npars = len(pars)
        lbounds,ubounds = bounds
        check = np.zeros(npars,np.int32)
        badlo, = np.where(pars<=lbounds)
        if len(badlo)>0:
            check[badlo] = 1
        badhi, = np.where(pars>=ubounds)
        if len(badhi):
            check[badhi] = 2
        return check
        
    def limbounds(self,pars,bounds):
        """ Limit the parameters to the boundaries."""
        lbounds,ubounds = bounds
        outpars = np.minimum(np.maximum(pars,lbounds),ubounds)
        return outpars

    def limsteps(self,steps,maxsteps):
        """ Limit the parameter steps to maximum step sizes."""
        signs = np.sign(steps)
        outsteps = np.minimum(np.abs(steps),maxsteps)
        outsteps *= signs
        return outsteps

    def steps(self,pars,dx=0.5):
        """ Return step sizes to use when fitting the stellar parameters."""
        npars = len(pars)
        fsteps = np.zeros(npars,float)
        fsteps[0:-1:3] = np.maximum(np.abs(pars[0:-1:3])*0.25,1)
        fsteps[1::3] = dx        
        fsteps[2::3] = dx
        fsteps[-1] = 10
        return fsteps
        
    def newpars(self,pars,steps,bounds,maxsteps):
        """ Get new parameters given initial parameters, steps and constraints."""

        # Limit the steps to maxsteps
        limited_steps = self.limsteps(steps,maxsteps)
        # Make sure that these don't cross the boundaries
        lbounds,ubounds = bounds
        check = self.checkbounds(pars+limited_steps,bounds)
        # Reduce step size for any parameters to go beyond the boundaries
        badpars = (check!=0)
        # reduce the step sizes until they are within bounds
        newsteps = limited_steps.copy()
        count = 0
        maxiter = 2
        while (np.sum(badpars)>0 and count<=maxiter):
            newsteps[badpars] /= 2
            newcheck = self.checkbounds(pars+newsteps,bounds)
            badpars = (newcheck!=0)
            count += 1
            
        # Final parameters
        newpars = pars + newsteps
            
        # Make sure to limit them to the boundaries
        check = self.checkbounds(newpars,bounds)
        badpars = (check!=0)
        if np.sum(badpars)>0:
            # add a tiny offset so it doesn't fit right on the boundary
            newpars = np.minimum(np.maximum(newpars,lbounds+1e-30),ubounds-1e-30)
        return newpars

    def psf(self,x,y,pars):
        """ Return the PSF model for a single star."""
        g,_ = mnb.psf(x,y,pars,self.psftype,self.psfparams,self.psflookup,self.imshape,
                      deriv=False,verbose=False)
        return g

    def psfjac(self,x,y,pars):
        """ Return the PSF model and derivatives/jacobian for a single star."""
        return mnb.psf(x,y,pars,self.psftype,self.psfparams,self.psflookup,self.imshape,
                      deriv=True,verbose=False)

    def starmodelfull(self,i):
        """ Calculate the model for a star for the full footprint."""
        # This is for a SINGLE star
        # Get all of the star data
        pars,xind,yind,ravelindex = self.stardata(i)
        n = len(xind)
        m = self.psf(xind,yind,pars)        
        return m

    def starmodel(self,i):
        """ Calculate the model for a star for the fitted pixels."""
        # This is for a SINGLE star
        # Get all of the star data
        pars,xind,yind,ravelindex = self.starfitdata(i)
        m = self.psf(xind,yind,pars)        
        return m
    
    def starjac(self,i):
        """ Calculate the jacobian for a star for the fitted pixels."""
        # This is for a SINGLE star
        # Get all of the star data
        pars,xind,yind,ravelindex = self.starfitdata(i)
        m,j = self.psfjac(xind,yind,pars)        
        return m,j

    def starfitim(self,i):
        """ Return the image pixels for a star."""
        n = self.starfit_ndata[i]
        ravelind = self.starfit_ravelindex[i,:n]
        return self.image.ravel()[ravelind]
    
    def starfiterr(self,i):
        """ Return the error pixels for a star."""
        n = self.starfit_ndata[i]
        ravelind = self.starfit_ravelindex[i,:n]
        return self.error.ravel()[ravelind]
    
    def starfitresid(self,i):
        """ Return the resid pixels for a star."""
        n = self.starfit_ndata[i]
        ravelind = self.starfit_ravelindex[i,:n]
        return self.resid[ravelind]
    
    # def starchisq(self,i,pars):
    #     """ Return chi-squared """
    #     fravelindex = self.starfitravelindex(i)
    #     resid = self.resid[fravelindex]
    #     # Subtract the sky?        
    #     err = self.error.ravel()[fravelindex]
    #     #flux = self.resid-self.skyflat
    #     #wt = 1/self.errflat**2
    #     wt = 1/wt**2
    #     bestmodel = self.model(pars,False,True)   # allparams,Trim
    #     resid = flux-bestmodel[self.usepix]
    #     chisq1 = np.sum(resid**2/err**2)
    #     return chisq1

        
    def starcov(self,i):
        """ Determine the covariance matrix for a single star."""

        # https://stats.stackexchange.com/questions/93316/parameter-uncertainty-after-non-linear-least-squares-estimation
        # more background here, too: http://ceres-solver.org/nnls_covariance.html

        # Get fitting pixel information
        pars,fxind,fyind,fravelindex = self.starfitdata(i)
        resid = self.resid[fravelindex]
        err = self.error.ravel()[fravelindex]
        wt = 1/err**2    # weights
        n = len(fxind)
        # Hessian = J.T * T, Hessian Matrix
        #  higher order terms are assumed to be small
        # https://www8.cs.umu.se/kurser/5DA001/HT07/lectures/lsq-handouts.pdf
        m,j = self.psfjac(fxind,fyind,pars)
        # Weights
        #   If weighted least-squares then
        #   J.T * W * J
        #   where W = I/sig_i**2
        hess = j.T @ (np.diag(wt) @ j)
        #hess = mjac.T @ mjac  # not weighted
        # cov = H-1, covariance matrix is inverse of Hessian matrix
        cov_orig = utils.inverse(hess)
        # Rescale to get an unbiased estimate
        # cov_scaled = cov * (RSS/(m-n)), where m=number of measurements, n=number of parameters
        # RSS = residual sum of squares
        #  using rss gives values consistent with what curve_fit returns
        #cov = cov_orig * (np.sum(resid**2)/(self.ntotpix-len(self.pars)))
        # Use chi-squared, since we are doing the weighted least-squares and weighted Hessian
        chisq = np.sum(resid**2/err**2)
        cov = cov_orig * (chisq/(n-3))  # what MPFITFUN suggests, but very small

        # cov = lqr.jac_covariange(mjac,resid,wt)
        
        return cov

    def starfit(self,i,minpercdiff=0.5):
        """
        Fit a single star
        single iteration of the non-linear least-squares loop
        """

        if i > self.nstars-1:
            raise IndexError('Index',i,' is out of bounds.',self.nstars,' stars')
        
        if self.freezestars[i]:
            return

        # Get fitting pixel information
        pars,fxind,fyind,fravelindex = self.starfitdata(i)
        resid = self.resid[fravelindex]
        err = self.error.ravel()[fravelindex]
        wt = 1/err**2    # weights

        # Calculate the local sky for this star (relative to the smooth sky image)
        sky = self.starsky(i)
        # subtract sky from the residuals
        resid -= sky
        
        # Jacobian solvers
        # Get the Jacobian and model
        #  only for pixels that are affected by the "free" parameters
        model0,j = self.psfjac(fxind,fyind,pars)
        # Solve Jacobian
        dbeta = utils.qr_jac_solve(j,resid,weight=wt)
        dbeta[~np.isfinite(dbeta)] = 0.0  # deal with NaNs, shouldn't happen
            
        # Perform line search
        # This has the previous best-fit model subtracted
        #  add it back in
        data = resid + model0
        # Models
        model1 = self.psf(fxind,fyind,pars+0.5*dbeta)
        model2 = self.psf(fxind,fyind,pars+dbeta)
        chisq0 = np.sum((data-model0)**2/err**2)
        chisq1 = np.sum((data-model1)**2/err**2)
        chisq2 = np.sum((data-model2)**2/err**2)
        if self.verbose:
            print('linesearch:',chisq0,chisq1,chisq2)
        alpha = utils.quadratic_bisector(np.array([0.0,0.5,1.0]),
                                         np.array([chisq0,chisq1,chisq2]))
        alpha = np.minimum(np.maximum(alpha,0.0),1.0)  # 0<alpha<1
        if np.isfinite(alpha)==False:
            alpha = 1.0
        pars_new = pars + alpha * dbeta
        new_dbeta = alpha * dbeta

        # Update parameters
        oldpars = pars.copy()
        bounds = self.mkbounds(pars,self.imshape)
        maxsteps = self.steps(pars)
        bestpars = self.newpars(pars,new_dbeta,bounds,maxsteps)
        # Check differences and changes
        diff = np.abs(bestpars-oldpars)
        percdiff = diff.copy()*0
        percdiff[0] = diff[0]/np.maximum(oldpars[0],0.0001)*100  # amp
        percdiff[1] = diff[1]*100               # x
        percdiff[2] = diff[2]*100               # y
        
        # Freeze parameters/stars that converged
        #  also subtract models of fixed stars
        #  also return new free parameters
        if self.nofreeze==False and np.sum(percdiff<=minpercdiff)==3:
            # frzpars = percdiff<=minpercdiff
            # freeparsind, = np.where(self.freezepars==False)
            # bestpar = self.freeze(bestpar,frzpars)

            # set NITER for this star

            self.freezestars[i] = True
            
            
            # # If the only free parameter is the sky offset,
            # #  then freeze it as well
            # if self.nfreepars==1 and self.freepars[-1]==True:
            #     self.freezepars[-1] = True
            #     bestpar = np.zeros((1),np.float64)
            # npar = len(bestpar)
            if self.verbose:
                print('Star ',i,' frozen')
        else:
            self.pars[3*i:3*i+3] = bestpars
        maxpercdiff = np.max(percdiff)
                
        # Update the residuals for a star's full footprint
        #  add in the previous full footprint model
        #  and subtract the new full footprint model
        _,xind,yind,ravelindex = self.stardata(i)        
        prevmodel = self.psf(xind,yind,oldpars)
        newmodel = self.psf(xind,yind,bestpars)
        self.resid[ravelindex] += prevmodel
        self.resid[ravelindex] -= newmodel

        # Update the model image
        self.modelim[ravelindex] += prevmodel
        self.modelim[ravelindex] -= newmodel
        
        # Calculate chisq with updated resid array
        bestchisq = np.sum(self.resid[fravelindex]**2/err**2)
        rms = np.sqrt(np.mean((self.resid[fravelindex]/bestpars[0])**2))
        if self.verbose:
            print('best chisq =',bestchisq)
        self.starchisq[i] = bestchisq
        self.starrms[i] = rms
        
        if self.verbose:
            print('Iter = ',self.niter)
            print('Pars = ',self.starpars[i])
            print('Percent diff = ',percdiff)
            print('Diff = ',diff)
            print('chisq = ',bestchisq)

    def chisq(self):
        """ Compute total chi-square of the current best-fit solution for all
        fitting pixels."""
        ravelindex = self.ravelindex    # all fitting pixels
        chisq = np.sum(self.resid[ravelindex]**2/self.error.ravel()[ravelindex]**2)
        return chisq
    
    def fit(self,maxiter=10,minpercdiff=0.5,reskyiter=2,nofreeze=False,verbose=False):
        """
        Fit PSF to all stars iteratively

        Parameters
        ----------
        maxiter : int, optional
           Maximum number of iterations to allow.  Only for methods "cholesky", "qr" or "svd".
           Default is 10.
        minpercdiff : float, optional
           Minimum percent change in the parameters to allow until the solution is
           considered converged and the iteration loop is stopped.  Only for methods
           "cholesky", "qr" and "svd".  Default is 0.5.
        reskyiter : int, optional
           After how many iterations to re-calculate the sky background. Default is 2.
        nofreeze : boolean, optional
           Do not freeze any parameters even if they have converged.  Default is False.
        verbose : boolean, optional
           Verbose output.

        Returns
        -------
        out : table
           Table of best-fitting parameters for each star.
           id, amp, amp_error, x, x_err, y, y_err, sky
        model : numpy array
           Best-fitting model of the stars and sky background.
        sky : numpy array
           Best-fitting sky image.

        Example
        -------

        outtab,model,sky = gf.fit()

        """

        # Iterate
        self.niter = 1
        maxpercdiff = 1e10
        while (self.niter<maxiter and self.nfreestars>0):
            start0 = clock()
            
            # Star loop
            for i in range(self.nstars):
                # Fit the single star (if not frozen)
                if self.freezestars[i]==False:
                    self.starfit(i)

            # Re-estimate the sky
            if self.niter % reskyiter == 0:
                if self.verbose:
                    print('Re-estimating the sky')
                self.sky()

            if verbose:
                print('iter dt =',(clock()-start0)/1e9,'sec.')
                
            self.niter += 1     # increment counter

            if self.verbose:
                print('niter=',self.niter)

        # Check that all starniter are set properly
        #  if we stopped "prematurely" then not all stars were frozen
        #  and didn't have starniter set
        self.starniter[np.where(self.starniter==0)] = self.niter

        # Calculate parameter uncertainties
        # estimate uncertainties
        # Calculate covariance matrix
        perror = np.zeros(len(self.pars),np.float64)
        for i in range(self.nstars):
            cov1 = self.starcov(i)
            perror1 = np.sqrt(np.diag(cov1))
            perror[3*i:3*i+3] = perror1
        self.perror[:] = perror

@njit(cache=True)
def numba_allfit(psftype,psfparams,psfnpix,psflookup,psfflux,
                 image,error,mask,tab,fitradius,maxiter=10,
                 minpercdiff=0.5,reskyiter=2,verbose=False,
                 nofreeze=False):
    """
    This is a thin wrapper around AllFitter()
    so we can cache it.
    """
    af = AllFitter(psftype,psfparams,psfnpix,psflookup,
                   image,error,mask,tab,fitradius,verbose,nofreeze)
    # Fit the stars iteratively
    #af.fit()
    af.fit(maxiter,minpercdiff,reskyiter,nofreeze,verbose)
    
    # Return the information that we want
    model = af.modelim.copy() #.reshape(af.imshape)
    skyim = af.skyim.copy()  #.reshape(af.imshape)
    pars = af.pars
    perror = af.perror
    # Put in catalog
    outtab = np.zeros((af.nstars,15),np.float64)
    outtab[:,0] = np.arange(af.nstars)+1           # id
    outtab[:,1] = pars[0:-1:3]                       # amp
    outtab[:,2] = perror[0:-1:3]                     # amp_error
    outtab[:,3] = pars[1::3]                         # x
    outtab[:,4] = perror[1::3]                       # x_error
    outtab[:,5] = pars[2::3]                         # y
    outtab[:,6] = perror[2::3]                       # y_error
    outtab[:,7] = af._starsky                        # sky
    outtab[:,8] = outtab[:,1]*psfflux                # flux
    outtab[:,9] = outtab[:,2]*psfflux                # flux_error
    outtab[:,10] = -2.5*np.log10(np.maximum(outtab[:,8],1e-10))+25.0   # mag
    outtab[:,11] = (2.5/np.log(10))*outtab[:,9]/outtab[:,8]            # mag_error
    outtab[:,12] = af.starrms
    outtab[:,13] = af.starchisq
    outtab[:,14] = af.starniter                      # niter, what iteration it converged on
    return outtab,model,skyim
    

def allfit(psf,image,tab,fitradius=0.0,maxiter=10,minpercdiff=0.5,
           reskyiter=2,nofreeze=False,verbose=False):
    """
    Fit PSF to all stars in an image iteratively.

    Parameters
    ----------
    psf : PSF object
       PSF object to use for the fitting.
    image : CCDData object
       Image to use to fit PSF model to stars.
    tab : table
       Catalog with initial amp/x/y values for the stars to use to fit the PSF.
       To pre-group the stars, add a "group_id" in the catalog.
    fitradius : float, optional
       The fitting radius in pixels.  By default the PSF FWHM is used.
           reskyiter=2,nofreeze=False,skyfit=True,verbose=False):
    maxiter : int, optional
       Maximum number of iterations to allow.  Only for methods "cholesky", "qr" or "svd".
       Default is 10.
    minpercdiff : float, optional
       Minimum percent change in the parameters to allow until the solution is
       considered converged and the iteration loop is stopped.  Only for methods
       "cholesky", "qr" and "svd".  Default is 0.5.
    reskyiter : int, optional
       After how many iterations to re-calculate the sky background. Default is 2.
    nofreeze : boolean, optional
       Do not freeze any parameters even if they have converged.  Default is False.
    verbose : boolean, optional
       Verbose output.

    Returns
    -------
    out : table
       Table of best-fitting parameters for each star.
       id, amp, amp_error, x, x_err, y, y_err, sky
    model : numpy array
       Best-fitting model of the stars and sky background.
    sky : numpy array
       Best-fitting sky image.

    Example
    -------

    outtab,model,sky = gf.fit()

    """

    t0 = time.time()

    # if skyfit==False:
    #     self.freezepars[-1] = True  # freeze sky value
        
    # # Centroids fixed
    # if recenter==False:
    #     self.freezepars[1::3] = True  # freeze X values
    #     self.freezepars[2::3] = True  # freeze Y values

    # Create AllFitter object
    fitradius = 3.0
    if psf.haslookup:
        psflookup = psf.lookup.copy().astype(np.float64)
    else:
        psflookup = np.zeros((1,1,1),np.float64)
    psfflux = psf.flux()
    out = numba_allfit(psf.psftype,psf.params,psf.npix,psflookup,psfflux,
                       image.data,image.error,image.mask,tab,fitradius,
                       maxiter,minpercdiff,reskyiter,verbose,nofreeze)
    outtab,model,skyim = out
    model = model.reshape(image.shape)
    skyim = skyim.reshape(image.shape)
    return outtab,model,skyim
        
    #af = AllFitter(psf.psftype,psf.params,psf.npix,psflookup,
    #               image.data,image.error,image.mask,tab,
    #               fitradius,verbose,nofreeze)
    ## Fit the stars iteratively
    #af.fit()
    
    # Make final model
    # self.unfreeze()
    model = af.modelim.copy().reshape(af.imshape)
    skyim = af.skyim.copy().reshape(af.imshape)

    # Parameters and uncertainties
    pars = af.pars
    perror = af.perror

    # Put in catalog
    outtab = np.zeros((af.nstars,15),np.float64)
    outtab[:,0] = np.arange(af.nstars)+1           # id
    outtab[:,1] = pars[0:-1:3]                       # amp
    outtab[:,2] = perror[0:-1:3]                     # amp_error
    outtab[:,3] = pars[1::3]                         # x
    outtab[:,4] = perror[1::3]                       # x_error
    outtab[:,5] = pars[2::3]                         # y
    outtab[:,6] = perror[2::3]                       # y_error
    outtab[:,7] = af._starsky                        # sky
    psfflux = psf.flux()
    outtab[:,8] = outtab[:,1]*psfflux                # flux
    outtab[:,9] = outtab[:,2]*psfflux                # flux_error
    outtab[:,10] = -2.5*np.log10(np.maximum(outtab[:,8],1e-10))+25.0   # mag
    outtab[:,11] = (2.5/np.log(10))*outtab[:,9]/outtab[:,8]            # mag_error
    outtab[:,12] = af.starrms
    outtab[:,13] = af.starchisq
    outtab[:,14] = af.starniter                      # niter, what iteration it converged on

    if verbose:
        print('dt =',(clock()-start)/1e9,'sec.')

    return outtab,model,skyim

        

#@njit
def fit(psf,image,tab,fitradius=0.0,recenter=True,maxiter=10,minpercdiff=0.5,
        reskyiter=2,nofreeze=False,skyfit=True,verbose=False):
    """
    Fit PSF to all stars in an image.

    To pre-group the stars, add a "group_id" in the input catalog.

    Parameters
    ----------
    psf : PSF object
       PSF object with initial parameters to use.
    image : CCDData object
       Image to use to fit PSF model to stars.
    tab : table
       Catalog with initial amp/x/y values for the stars to use to fit the PSF.
       To pre-group the stars, add a "group_id" in the catalog.
    fitradius : float, optional
       The fitting radius in pixels.  By default the PSF FWHM is used.
    recenter : boolean, optional
       Allow the centroids to be fit.  Default is True.
    maxiter : int, optional
       Maximum number of iterations to allow.  Only for methods "qr" or "svd".
       Default is 10.
    minpercdiff : float, optional
       Minimum percent change in the parameters to allow until the solution is
       considered converged and the iteration loop is stopped.  Only for methods
       "qr" and "svd".  Default is 0.5.
    reskyiter : int, optional
       After how many iterations to re-calculate the sky background. Default is 2.
    nofreeze : boolean, optional
       Do not freeze any parameters even if they have converged.  Default is False.
    skyfit : boolean, optional
       Fit a constant sky offset with the stellar parameters.  Default is True.
    verbose : boolean, optional
       Verbose output.

    Returns
    -------
    results : table
       Table of best-fitting parameters for each star.
       id, amp, amp_error, x, x_err, y, y_err, sky
    model : numpy array
       Best-fitting model of the stars and sky background.

    Example
    -------

    results,model = fit(psf,image,tab,groups)

    """

    start = clock()
    nstars = len(tab)
    ny,nx = image.shape

    # Groups
    if 'group_id' not in tab.keys():
        daogroup = DAOGroup(crit_separation=2.5*psf.fwhm())
        starlist = tab.copy()
        starlist['x_0'] = tab['x']
        starlist['y_0'] = tab['y']
        # THIS TAKES ~4 SECONDS!!!!!! WAY TOO LONG!!!!
        star_groups = daogroup(starlist)
        tab['group_id'] = star_groups['group_id']

    # Star index
    starindex = utils.index(np.array(tab['group_id']))
    #starindex = dln.create_index(np.array(tab['group_id'])) 
    groups = starindex['value']
    ngroups = len(groups)
    if verbose:
        print(ngroups,'star groups')

    # Initialize catalog
    #dt = np.dtype([('id',int),('amp',float),('amp_error',float),('x',float),
    #               ('x_error',float),('y',float),('y_error',float),('sky',float),
    #               ('flux',float),('flux_error',float),('mag',float),('mag_error',float),
    #               ('niter',int),('group_id',int),('ngroup',int),('rms',float),('chisq',float)])
    outtab = np.zeros((nstars,17),dtype=dt)
    outtab[:,0] = tab[:,0]  # copy ID


    # Group Loop
    #---------------
    resid = image.copy()
    outmodel = np.zeros(image.shape,np.float64)
    outsky = np.zeros(image.shape,np.float64)
    for g,grp in enumerate(groups):
        ind = starindex['index'][starindex['lo'][g]:starindex['hi'][g]+1]
        nind = len(ind)
        inptab = tab[ind].copy()
        if 'amp' not in inptab.columns:
            # Estimate amp from flux and fwhm
            # area under 2D Gaussian is 2*pi*A*sigx*sigy
            if 'fwhm' in inptab.columns:
                amp = inptab['flux']/(2*np.pi*(inptab['fwhm']/2.35)**2)
            else:
                amp = inptab['flux']/(2*np.pi*(psf.fwhm()/2.35)**2)                
            staramp = np.maximum(amp,0)   # make sure it's positive
            inptab['amp'] = staramp
        
        if verbose:
            print('-- Group '+str(grp)+'/'+str(len(groups))+' : '+str(nind)+' star(s) --')

        # Single Star
        if nind==1:
            inptab = np.array([inptab[1],inptab[2],inptab[3]])
            out,model = psf.fit(resid,inptab,niter=3,verbose=verbose,retfullmodel=True,recenter=recenter)
            model.data -= out['sky']   # remove sky
            outmodel.data[model.bbox.slices] += model.data
            outsky.data[model.bbox.slices] = out['sky']

        # Group
        else:
            bbox = cutoutbbox(image,psf,inptab)
            out,model,sky = gfit.fit(psf,resid[bbox.slices],inptab,fitradius=fitradius,
                                     recenter=recenter,maxiter=maxiter,minpercdiff=minpercdiff,
                                     reskyiter=reskyiter,nofreeze=nofreeze,verbose=verbose,
                                     skyfit=skyfit,absolute=True)
            outmodel.data[model.bbox.slices] += model.data
            outsky.data[model.bbox.slices] = sky
            
        # Subtract the best model for the group/star
        resid[model.bbox.slices].data -= model.data

        # Put in catalog
        cols = ['amp','amp_error','x','x_error','y','y_error',
                'sky','flux','flux_error','mag','mag_error','niter','rms','chisq']
        for c in cols:
            outtab[c][ind] = out[c]
        outtab['group_id'][ind] = grp
        outtab['ngroup'][ind] = nind
        outtab = Table(outtab)
        
    if verbose:
        print('dt = {:.2f} sec.'.format(time.time()-t0))
    
    return outtab,outmodel,outsky
        
            



import os
import numpy as np
from numba import njit,types,from_dtype
from numba.experimental import jitclass
from numba_kdtree import KDTree
from . import models_numba as mnb, utils_numba as utils

# Fit a PSF model to multiple stars in an image

@njit
def starcube(tab,image,error,npix=51,fillvalue=np.nan):
    """
    Produce a cube of cutouts of stars.

    Parameters
    ----------
    tab : table
       The catalog of stars to use.  This should have "x" and "y" columns and
         preferably also "amp".
    image : CCDData object
       The image to use to generate the stellar images.
    fillvalue : float, optional
       The fill value to use for pixels that are bad are off the image.
            Default is np.nan.

    Returns
    -------
    cube : numpy array
       Two-dimensional cube (Npix,Npix,Nstars) of the star images.

    Example
    -------

    cube = starcube(tab,image)

    """

    # Get the residuals data
    nstars = len(tab)
    nhpix = npix//2
    cube = np.zeros((npix,npix,nstars),float)
    xx,yy = utils.meshgrid(np.arange(npix)-nhpix,np.arange(npix)-nhpix)
    rr = np.sqrt(xx**2+yy**2)        
    x = xx[0,:]
    y = yy[:,0]
    for i in range(nstars):
        #xcen = tab['x'][i]
        #ycen = tab['y'][i]
        xcen = tab[1][i]
        ycen = tab[2][i]
        bbox = mnb.starbbox((xcen,ycen),image.shape,nhpix)
        flux = bbox.slice(image)
        err = bbox.slice(error)
        #amp = tab['peak'][i]
        amp = tab[0][i]
        xim,yim = bbox.xy()
        xim = xim-xcen
        yim = yim-ycen
        # Need to interpolate onto a regular grid
        im2 = np.zeros((npix,npix),float)+np.nan
        xind, = np.where((x>=bbox.ixmin-xcen) & (x<=bbox.ixmax-1-xcen))
        xmin,xmax = np.min(xind),np.max(xind)
        yind, = np.where((y>=bbox.iymin-ycen) & (y<=bbox.iymax-1-ycen))
        ymin,ymax = np.min(yind),np.max(yind)
        # xx/yy are relative coordinates for the final interpolated image
        #       i.e. [-3, -2, -1, 0, 1, 2, 3]
        # xim/yim are relative coordinates for the original flux image
        # for the interpolation we need coordinates relative to to (0,0) of the
        #    flux image
        fnx = xmax-xmin+1
        fny = ymax-ymin+1
        fxx = xx[ymin:ymax+1,xmin:xmax+1]-xim[0,0]   # relative to flux origin
        fyy = yy[ymin:ymax+1,xmin:xmax+1]-yim[0,0]
        fim = utils.alinearinterp(flux,fxx,fyy)
        fim = fim.reshape((fny,fnx))
        im2[ymin:ymax+1,xmin:xmax+1] = fim
        # Stuff it into 3D array
        cube[:,:,i] = im2
    return cube


@njit
def mkempirical(cube,order=0,coords=None,shape=None,lookup=False):
    """
    Take a star cube and collapse it to make an empirical PSF using median
    and outlier rejection.

    Parameters
    ----------
    cube : numpy array
      Three-dimensional cube of star images (or residual images) of shape
        (Npix,Npix,Nstars).
    order : int, optional
      The order of the variations. 0-constant, 1-linear terms.  If order=1,
        Then coords and shape must be input.
    coords : tuple, optional
      Two-element tuple of the X/Y coordinates of the stars.  This is needed
        to generate the linear empirical model (order=1).
    shape : tuple, optional
      Two-element tuple giving the shape (Ny,Nx) of the image.  This is
        needed to generate the linear empirical model (order=1).
    lookup : boolean, optional
      Parameter to indicate if this is a lookup table.  If lookup=False, then
      the constant term is constrained to be non-negative. Default is False.

    Returns
    -------
    epsf : numpy array
      The empirical PSF model  If order=0, then this is just a 2D image.  If
        order=1, then it will be a 3D cube (Npix,Npix,4) where the four terms
        are [constant, X-term, Y-term, X*Y-term].  If rect=True, then a list
        of RectBivariateSpline functions are returned.

    Example
    -------

    epsf = mkempirical(cube,order=0)

    or

    epsf = mkempirical(cube,order=1,coords=coords,shape=im.shape)

    """

    ny,nx,nstar = cube.shape
    npix = ny
    nhpix = ny//2
    
    # Do outlier rejection in each pixel
    med = utils.median3d(cube,axis=2,ignore_nan=True)
    bady,badx = np.where(np.isfinite(med)==False)
    if len(bady)>0:
        gdmed = utils.median(med,ignore_nan=True)
        for i in range(len(bady)):
            med[bady[i],badx[i]] = gdmed
    sig = utils.mad3d(cube,axis=2,ignore_nan=True)
    bady,badx = np.where(np.isfinite(sig)==False)
    if len(bady)>0:
        gdsig = utils.median(sig,ignore_nan=True)
        for i in range(len(bady)):
            sig[bady[i],badx[i]] = gdsig
    # Mask outlier points
    bady,badx,badk = np.where((np.abs(cube-med.reshape((med.shape)+(-1,)))>3*sig.reshape((med.shape)+(-1,)))
                              & np.isfinite(cube))
    if len(bady)>0:
        macube = cube.copy()
        for i in range(len(bady)):
            macube[bady[i],badx[i],badk[i]] = np.nan
        medim = utils.mean3d(macube,axis=2,ignore_nan=True)
        # Remove any final NaNs
        bady,badx = np.where(np.isfinite(medim)==False)
        if len(bady)>0:
            gdmed = utils.median(medim,ignore_nan=True)
            for i in range(len(bady)):
                medim[bady[i],badx[i]] = gdmed
    else:
        medim = med

    # Check how well each star fits the median
    goodpix = np.sum(np.sum(np.isfinite(macube),axis=0),axis=0)
    print('goodpix=',goodpix)
    #goodpix = macube.count(axis=(0,1))
    resid2 = (cube-medim.reshape((medim.shape)+(-1,)))**2
    mnresid2 = utils.sum2d(utils.sum3d(resid2,axis=0,ignore_nan=True),axis=0,ignore_nan=True)/goodpix
    rms = np.sqrt(mnresid2)
    print('rms=',rms)
    #print(utils.sum3d(dum,axis=0,ignore_nan=True))
    #rms = np.sqrt(np.nansum((cube-medim.reshape((medim.shape)+(-1,)))**2,axis=(0,1))/goodpix)

    xx,yy = utils.meshgrid(np.arange(npix)-nhpix,np.arange(npix)-nhpix)
    rr = np.sqrt(xx**2+yy**2)        
    x = xx[0,:]
    y = yy[:,0]
    mask = (rr<=nhpix)
    
    # Constant
    if order==0:
        # Make sure it goes to zero at large radius
        medim *= mask  # mask corners
        # Make sure values are positive
        if lookup==False:
            medim = np.maximum(medim,0.0)
            fpars = medim
        else:
            fpars = medim
        fpars = fpars.reshape((ny,nx,1))
            
    # Linear
    elif order==1:
        if coords is None or shape is None:
            raise ValueError('Need coords and shape with order=1')
        pars = np.zeros((ny,nx,4),float)
        # scale coordinates to -1 to +1
        xcen,ycen = coords[:,0],coords[:,1]
        relx,rely = mnb.relcoord(xcen,ycen,shape)
        # Loop over pixels and fit line to x/y
        for i in range(ny):
            for j in range(nx):
                data1 = cube[i,j,:]
                if np.sum(np.abs(data1)) != 0:
                    # maybe use a small maxiter
                    pars1,perror1,pcov1 = utils.poly2dfit(relx,rely,data1,data1*0+1)
                    pars[i,j,:] = pars1
        # Make sure it goes to zero at large radius
        fpars = pars.copy()
        for i in range(4):
            # Make sure edges are zero on average for higher-order terms
            gdouter, = np.where(rr.ravel()>nhpix*0.8)
            medouter = np.median(pars[:,:,i].ravel()[gdouter])
            fpars[:,:,i] -= medouter
            # Mask corners
            fpars[:,:,i] *= mask
            # Each higher-order term must have ZERO total volume
   
    return fpars,rms

#@njit
def findpsfnei(alltab,psftab,npix):
    """
    Find stars near PSF stars.

    Parameters
    ----------
    alltab : numpy array
      Catalog of all sources in the image.
    psftab : numpy array
      Catalog of PSF stars.
    npix : int
      Search radius in pixels.

    Returns
    -------
    indall : numpy array
       List of indices into allcat that are neighbors to
        the PSF stars (within radius of npix), and are not
        PSF stars themselves.

    Example
    -------

    indall = findpsfnei(alltab,psftab,npix)

    """
    # Returns distance and index of closest neighbor
    
    nalltab = len(alltab)
    npsftab = len(psftab)
    
    # Use KD-tree
    x1 = np.zeros((nalltab,2),float)
    x1[:,0] = alltab[:,2]
    x1[:,1] = alltab[:,3]
    x2 = np.zeros((npsftab,2),float)
    x2[:,0] = psftab[:,2]
    x2[:,1] = psftab[:,3]


    #data = np.random.random(3_000_000).reshape(-1, 3)
    #kdtree = KDTree(data, leafsize=10)

    ## query the nearest neighbors of the first 100 points
    #distances, indices, neighbors = kdtree.query(data[:100], k=30)

    ## query all points in a radius around the first 100 points
    #indices = kdtree.query_radius(data[:100], r=0.5, return_sorted=True)

    
    #kdt = cKDTree(x1)
    kdt = KDTree(x1)
    # Get distance for 2 closest neighbors
    dist, ind = kdt.query(x2, k=50, distance_upper_bound=np.sqrt(2)*npix//2)
    # closest neighbor is always itself, remove it 
    dist = dist[:,1:]
    ind = ind[:,1:]
    # Add up all the stars
    gdall, = np.where(np.isfinite(dist.ravel()))
    indall = ind.ravel()[gdall]
    # Get unique ones
    indall = np.unique(indall)

    # Make sure none of them are our psf stars
    ind1,ind2,dist = utils.xmatch(alltab[indall,2],alltab[indall3],
                                  psftab[:,2],psftab[:,3],5,sphere=False)
    #    alltab['x'][indall],alltab['y'][indall],psftab['x'],psftab['y'],5)
    # Remove any matches
    if len(ind1)>0:
        indall = np.delete(indall,ind1)
    
    return indall
    
#@njit
def subtractnei(image,error,alltab,psftab,psf):
    """
    Subtract neighboring stars to PSF stars from the image.

    Parameters
    ----------
    image : CCDDdata object
       The input image from which to subtract PSF neighbor stars.
    alltab : numpy array
      Catalog of all sources in the image.
    psftab : numpy array
      Catalog of PSF stars.
    psf : PSF object
      The PSF model.

    Returns
    -------
    resid : CCDData object
      The input images with the PSF neighbor stars subtracted.

    Example
    -------

    subim = subtractnei(image,alltab,psftab,psf)

    """

    indnei = findpsfnei(alltab,psftab,psf.npix)
    nnei = len(indnei)

    flux = image
    resid = image.copy()
    fitradius = psf.fwhm()*0.5
    
    # Loop over neighboring stars and fit just the core
    for i in range(nnei):
        x1 = alltab[indnei[i],2]
        xp1 = int(np.minimum(np.maximum(np.round(x1),0),image.shape[1]-1))
        y1 = alltab[indnei[i],3]
        yp1 = int(np.minimum(np.maximum(np.round(y1),0),image.shape[0]-1))
        h1 = alltab[indnei[i],1]
        initpars = np.array([h1,x1,y1])
        bbox = starbbox((initpars[1],initpars[2]),image.shape,psf.radius)
        # Fit amp empirically with central pixels
        flux1 = bbox.slice(flux)
        err1 = bbox.slice(error)
        model1 = psf(pars=initpars,bbox=bbox)
        good = ((flux1/err1>2) & (flux1>0) & (model1/np.max(model1)>0.25))
        amp = np.median(flux1[good]/model1[good]) * initpars[0]
        pars = np.array([amp, x1, y1])
        im1 = psf(pars=pars,bbox=bbox)
        resid[bbox.slices].data -= im1
    return resid

@njit
def starbbox(coords,imshape,radius):
    """                                                                                         
    Return the boundary box for a star given radius and image size.                             
                                                                                                
    Parameters                                                                                  
    ----------                                                                                  
    coords: list or tuple                                                                       
       Central coordinates (xcen,ycen) of star (*absolute* values).                             
    imshape: list or tuple                                                                      
       Image shape (ny,nx) values.  Python images are (Y,X).                                    
    radius: float                                                                               
       Radius in pixels.                                                                        
                                                                                                
    Returns                                                                                     
    -------                                                                                     
    bbox : BoundingBox object                                                                   
       Bounding box of the x/y ranges.                                                          
       Upper values are EXCLUSIVE following the python convention.                              
                                                                                                
    """

    # pixels span +/-0.5 pixels in each direction
    # so pixel x=5 spans x=4.5-5.5
    # Use round() to get the value of the pixel that is covers
    # the coordinate
    # to include pixels at the bottom but not the top we have to
    # subtract a tiny bit before we round
    eta = 1e-5

    # Star coordinates
    xcen,ycen = coords
    ny,nx = imshape   # python images are (Y,X)
    x0 = xcen-radius
    x1 = xcen+radius
    y0 = ycen-radius
    y1 = ycen+radius
    xlo = np.maximum(int(np.round(x0-eta)),0)
    xhi = np.minimum(int(np.round(x1-eta))+1,nx)
    ylo = np.maximum(int(np.round(y0-eta)),0)
    yhi = np.minimum(int(np.round(y1-eta))+1,ny)
    # add 1 at the upper end because those values are EXCLUDED
    # by the standard python convention

    # The old way of doing it
    #xlo = np.maximum(int(np.floor(xcen-radius)),0)
    #xhi = np.minimum(int(np.ceil(xcen+radius+1)),nx)
    #ylo = np.maximum(int(np.floor(ycen-radius)),0)
    #yhi = np.minimum(int(np.ceil(ycen+radius+1)),ny)
    return np.array([xlo,xhi,ylo,yhi])

# @njit
# def sliceinsert(array,lo,insert):
#     """ Insert array values."""
#     n = insert.size
#     for i in range(n):
#         j = i + lo
#         array[j] = insert[i]
#     # array is updated in place  


@njit
def getstar(image,error,xcen,ycen,fitradius):
    """ Return the entire footprint image/error/x/y arrays for one star."""
    # always return the same size
    # a distance of 6.2 pixels spans 6 full pixels but you could have
    # 0.1 left on one side and 0.1 left on the other side
    # that's why we have to add 2 pixels
    npix = int(np.floor(2*fitradius))+2
    bbox = starbbox((xcen,ycen),image.shape,fitradius)
    nx = bbox[1]-bbox[0]
    ny = bbox[3]-bbox[2]
    # extra buffer is ALWAYS at the end of each dimension    
    imdata = np.zeros(npix*npix,float)+np.nan
    errdata = np.zeros(npix*npix,float)+np.nan
    xdata = np.zeros(npix*npix,np.int32)-1
    ydata = np.zeros(npix*npix,np.int32)-1
    count = 0
    for j in range(npix):
        y = j + bbox[2]
        for i in range(npix):
            x = i + bbox[0]            
            xdata[count] = x
            ydata[count] = y
            if x>=bbox[0] and x<=bbox[1]-1 and y>=bbox[2] and y<=bbox[3]-1:
                imdata[count] = image[y,x]
                errdata[count] = error[y,x]
            count += 1
    return imdata,errdata,xdata,ydata,bbox,nx,ny

@njit
def collatestars(image,error,starx,stary,fitradius):
    """ Get the entire footprint image/error/x/y for all of the stars."""
    nstars = len(starx)
    npix = int(np.floor(2*fitradius))+2
    # Get xdata, ydata, error
    imdata = np.zeros((nstars,npix*npix),float)
    errdata = np.zeros((nstars,npix*npix),float)
    xdata = np.zeros((nstars,npix*npix),np.int32)
    ydata = np.zeros((nstars,npix*npix),np.int32)
    bbox = np.zeros((nstars,4),np.int32)
    shape = np.zeros((nstars,2),np.int32)
    for i in range(nstars):
        imdata1,errdata1,xdata1,ydata1,bbox1,nx1,ny1 = getstar(image,error,starx[i],stary[i],fitradius)
        imdata[i,:] = imdata1
        errdata[i,:] = errdata1
        xdata[i,:] = xdata1
        ydata[i,:] = ydata1
        bbox[i,:] = bbox1
        shape[i,0] = ny1
        shape[i,1] = nx1
    return imdata,errdata,xdata,ydata,bbox,shape

@njit
def unpackstar(imdata,errdata,xdata,ydata,bbox,shape,istar):
    """ Return unpacked data for one star."""
    imdata1 = imdata[istar,:].copy()
    errdata1 = errdata[istar,:].copy()
    xdata1 = xdata[istar,:].copy()
    ydata1 = ydata[istar,:].copy()
    bbox1 = bbox[istar,:]
    shape1 = shape[istar,:]
    n = len(imdata1)
    npix = int(np.sqrt(n))
    # Convert to 2D arrays
    imdata1 = imdata1.reshape(npix,npix)
    errdata1 = errdata1.reshape(npix,npix)
    xdata1 = xdata1.reshape(npix,npix)
    ydata1 = ydata1.reshape(npix,npix)
    # Trim values
    if shape1[0] < npix or shape1[1] < npix:
        imdata1 = imdata1[:shape1[0],:shape1[1]]
        errdata1 = errdata1[:shape1[0],:shape1[1]]
        xdata1 = xdata1[:shape1[0],:shape1[1]]
        ydata1 = ydata1[:shape1[0],:shape1[1]]
    return imdata1,errdata1,xdata1,ydata1,bbox1,shape1

@njit
def getfitstar(image,error,xcen,ycen,fitradius):
    """ Get the fitting pixels for a single star."""
    npix = int(np.floor(2*fitradius))+2
    bbox = starbbox((xcen,ycen),image.shape,fitradius)
    flux = image[bbox[2]:bbox[3],bbox[0]:bbox[1]].copy()
    err = error[bbox[2]:bbox[3],bbox[0]:bbox[1]].copy()
    nflux = flux.size
    nx = bbox[1]-bbox[0]
    ny = bbox[3]-bbox[2]
    imdata = np.zeros(npix*npix,float)
    errdata = np.ones(npix*npix,float)
    xdata = np.zeros(npix*npix,np.int32)-1
    ydata = np.zeros(npix*npix,np.int32)-1
    mask = np.zeros(npix*npix,np.int32)
    count = 0
    for j in range(ny):
        y = j + bbox[2]
        for i in range(nx):
            x = i + bbox[0]
            r = np.sqrt((x-xcen)**2 + (y-ycen)**2)
            if r <= fitradius:
                imdata[count] = flux[j,i]
                errdata[count] = error[j,i]
                xdata[count] = x
                ydata[count] = y
                mask[count] = 1
                count += 1
    return imdata,errdata,xdata,ydata,count,mask
        
@njit
def collatefitstars(image,error,starx,stary,fitradius):
    nstars = len(starx)
    npix = int(np.floor(2*fitradius))+2
    # Get xdata, ydata, error
    maxpix = nstars*(npix)**2
    imdata = np.zeros((nstars,npix*npix),float)
    errdata = np.zeros((nstars,npix*npix),float)
    xdata = np.zeros((nstars,npix*npix),np.int32)
    ydata = np.zeros((nstars,npix*npix),np.int32)
    mask = np.zeros((nstars,npix*npix),np.int32)
    ndata = np.zeros(nstars,np.int32)
    bbox = np.zeros((nstars,4),np.int32)
    for i in range(nstars):
        xcen = starx[i]
        ycen = stary[i]
        bb = starbbox((xcen,ycen),image.shape,fitradius)
        imdata1,errdata1,xdata1,ydata1,n1,mask1 = getfitstar(image,error,xcen,ycen,fitradius)
        imdata[i,:] = imdata1
        errdata[i,:] = errdata1
        xdata[i,:] = xdata1
        ydata[i,:] = ydata1
        mask[i,:] = mask1
        ndata[i] = n1
        bbox[i,:] = bb
    return imdata,errdata,xdata,ydata,bbox,ndata,mask

@njit
def unpackfitstar(imdata,errdata,xdata,ydata,bbox,ndata,istar):
    """ Return unpacked fitting data for one star."""
    imdata1 = imdata[istar,:]
    errdata1 = errdata[istar,:]
    xdata1 = xdata[istar,:]
    ydata1 = ydata[istar,:]
    bbox1 = bbox[istar,:]
    n1 = ndata[istar]
    # Trim to the values we want
    imdata1 = imdata1[:n1]
    errdata1 = errdata1[:n1]
    xdata1 = xdata1[:n1]
    ydata1 = ydata1[:n1]
    return imdata1,errdata1,xdata1,ydata1,bbox1,n1

    
kv_ty = (types.int64, types.unicode_type)
spec = [
    ('psftype', types.int32),
    ('params', types.float64[:]),
    #('lookup', types.float64[:,:,:]),
    #('order', types.int32),
    #('fwhm', types.float64),
    ('image', types.float64[:,:]),
    ('error', types.float64[:,:]),
    ('starinit', types.float64[:,:]),
    ('nstars', types.int32),
    ('niter', types.int32),
    ('npsfpix', types.int32),
    ('nx', types.int32),
    ('ny', types.int32),
    ('fitradius', types.float64),
    ('nfitpix', types.int32),
    ('staramp', types.float64[:]),
    ('starxcen', types.float64[:]),
    ('starycen', types.float64[:]),
    ('starchisq', types.float64[:]),
    ('starrms', types.float64[:]),
    ('starnpix', types.int32[:]),
    ('star_imdata', types.float64[:,:]),
    ('star_errdata', types.float64[:,:]),
    ('star_xdata', types.int32[:,:]),
    ('star_ydata', types.int32[:,:]),
    ('star_bbox', types.int32[:,:]),
    ('star_shape', types.int32[:,:]),
    ('starfit_imdata', types.float64[:,:]),
    ('starfit_errdata', types.float64[:,:]),
    ('starfit_xdata', types.int32[:,:]),
    ('starfit_ydata', types.int32[:,:]),
    ('starfit_mask', types.int32[:,:]),
    ('starfit_bbox', types.int32[:,:]),
    ('starfit_ndata', types.int32[:]),
    ('npix', types.int32),
    ('radius', types.int32),
    ('verbose', types.boolean),
    ('niter', types.int32),
    ('_unitfootflux', types.float64),
    #('lookup', types.float64[:,:,:]),
    #('_bounds', types.float64[:,:]),
    #('_steps', types.float64[:]),
    ('coords', types.float64[:]),
    ('imshape', types.int32[:]),
    #('order', types.int32),
]

@jitclass(spec)
class PSFFitter(object):

    def __init__(self,psftype,psfparams,image,error,starx,stary,starflux,fitradius,verbose):
        self.verbose = verbose
        self.psftype = psftype
        self.params = psfparams
        #self.lookup = psf.lookup
        #self.order = psf.order
        #self.fwhm = psf.fwhm()
        self.image = image.astype(np.float64)
        self.error = error.astype(np.float64)
        self.imshape = np.zeros(2,np.int32)
        self.imshape[:] = image.shape
        nstars = len(starx)
        self.starinit = np.zeros((nstars,3),float)
        self.starinit[:,0] = starx
        self.starinit[:,1] = stary
        self.starinit[:,2] = starflux
        self.nstars = nstars
        self.niter = 0
        #self.npsfpix = psf.npix
        ny,nx = image.shape
        self.nx = nx
        self.ny = ny
        #if np.isfinite(fitradius)==False:
        #    fitradius = self.fwhm*1.5
        self.fitradius = fitradius
        self.nfitpix = int(np.floor(2*fitradius))+2  # max pixels 
        self.staramp = np.zeros(self.nstars,float)
        self.staramp[:] = starflux/(2*fitradius+1)**2
        # current best-fit values
        self.starxcen = np.zeros(self.nstars,float)
        self.starxcen[:] = self.starinit[:,0].copy()
        self.starycen = np.zeros(self.nstars,float)
        self.starycen[:] = self.starinit[:,1].copy()
        self.starchisq = np.zeros(self.nstars,float)
        self.starrms = np.zeros(self.nstars,float)
        self.starnpix = np.zeros(self.nstars,np.int32)

        # Get information for all the stars
        imdata,errdata,xdata,ydata,bbox,shape = collatestars(image,error,self.starxcen,
                                                             self.starycen,fitradius)
        self.star_imdata = imdata
        self.star_errdata = errdata
        self.star_xdata = xdata
        self.star_ydata = ydata
        self.star_bbox = bbox
        self.star_shape = shape
        
        # Get fitting information for all the stars
        fimdata,ferrdata,fxdata,fydata,fbbox,fndata,fmask = collatefitstars(image,error,self.starxcen,
                                                                            self.starycen,fitradius)
        self.starfit_imdata = fimdata
        self.starfit_errdata = ferrdata
        self.starfit_xdata = fxdata
        self.starfit_ydata = fydata
        self.starfit_mask = fmask
        self.starfit_bbox = fbbox
        self.starfit_ndata = fndata

    def unpackstar(self,istar):
        return unpackstar(self.star_imdata,self.star_errdata,self.star_xdata,
                          self.star_ydata,self.star_bbox,self.star_shape,istar)

    def unpackfitstar(self,istar):
        return unpackfitstar(self.starfit_imdata,self.starfit_errdata,self.starfit_xdata,
                             self.starfit_ydata,self.starfit_bbox,self.starfit_ndata,istar)

    def psf(self,x,y,pars,psfparams,lookup,deriv):
        """ Get a PSF model for a single star."""
        return mnb.psf(x,y,pars,self.psftype,psfparams,lookup,
                       self.imshape,deriv=deriv,verbose=self.verbose)

    def model(self,x,args):
        """ model function."""
        # input the model parameters
        
        if self.verbose:
            print('model: '+str(self.niter)+' ',args)
        nparsarr = np.array([6,7,8,8,7])
        npars = nparsarr[self.psftype-1]
        # Loop over the stars and generate the model image
        allmodel = np.zeros(self.star_imdata.shape,float)
        maxpix = self.star_imdata.shape[1]
        for i in range(self.nstars):
            im1,err1,xdata1,ydata1,bbox1,n1 = self.unpackfitstar(i)
            allpars = np.zeros(npars,float)
            allpars[0] = self.staramp[i]
            allpars[1] = self.starxcen[i]
            allpars[2] = self.starycen[i]
            allpars[3:] = args
            g,_ = mnb.amodel2d(xdata1,ydata1,self.psftype,allpars,0)
            allmodel[i,:len(g)] = g
        return allmodel

    def chisq(self,pars):
        """ Create the model and return chisq for this set of PSF parameters."""
        m = self.model(np.arange(10),pars)
        chisq = np.sum((self.starfit_imdata-m)**2/self.starfit_errdata**2)        
        return chisq
        
    
            #image = self.imdata[i]
            #amp = self.staramp[i]
            #xcenorig = self.starxcenorig[i]   
            #ycenorig = self.starycenorig[i]
            #xcen = self.starxcen[i]   
            #ycen = self.starycen[i]            
            #bbox = self.bboxdata[i]
            #x = self.xlist[i]
            #y = self.ylist[i]
            #pixstart = self.pixstart[i]
            #npix = self.npix[i]
            #flux = self.imflatten[pixstart:pixstart+npix]
            #err = self.errflatten[pixstart:pixstart+npix]

            #x0orig = xcenorig - bbox.ixmin
            #y0orig = ycenorig - bbox.iymin
            #x0 = xcen - bbox.ixmin
            #y0 = ycen - bbox.iymin            
            
            # # Fit amp/xcen/ycen if niter=1
            # if refit:
            #     #if (self.niter<=1): # or self.niter%3==0):
            #     if self.niter>-1:
            #         # force the positions to stay within +/-2 pixels of the original values
            #         bounds = (np.array([0,np.maximum(x0orig-2,0),np.maximum(y0orig-2,0),-np.inf]),
            #                   np.array([np.inf,np.minimum(x0orig+2,bbox.shape[1]-1),np.minimum(y0orig+2,bbox.shape[0]-1),np.inf]))
            #         # the image still has sky in it, use sky (nosky=False)
            #         if np.isfinite(psf.fwhm())==False:
            #             print('nan fwhm')
            #             #import pdb; pdb.set_trace()
            #         pars,perror,model = psf.fit(image,[amp,x0,y0],nosky=False,retpararray=True,niter=5,bounds=bounds)
            #         xcen += (pars[1]-x0)
            #         ycen += (pars[2]-y0)
            #         amp = pars[0]                    
            #         self.staramp[i] = amp
            #         self.starxcen[i] = xcen
            #         self.starycen[i] = ycen
            #         model = psf(x,y,pars=[amp,xcen,ycen])
            #         if verbose:
            #             print('Star '+str(i)+' Refitting all parameters')
            #             print(str([amp,xcen,ycen]))

            #         #pars2,model2,mpars2 = psf.fit(image,[amp,x0,y0],nosky=False,niter=5,allpars=True)
            #         #import pdb; pdb.set_trace()
                        
    #             # Only fit amp if niter>1
    #             #   do it empirically
    #             else:
    #                 #im1 = psf(pars=[1.0,xcen,ycen],bbox=bbox)
    #                 #wt = 1/image.error**2
    #                 #amp = np.median(image.data[mask]/im1[mask])                
    #                 model1 = psf(x,y,pars=[1.0,xcen,ycen])
    #                 wt = 1/err**2
    #                 amp = np.median(flux/model1)
    #                 #amp = np.median(wt*flux/model1)/np.median(wt)

    #                 self.staramp[i] = amp
    #                 model = model1*amp
    #                 #self.starxcen[i] = pars2[1]+xy[0][0]
    #                 #self.starycen[i] = pars2[2]+xy[1][0]       
    #                 #print(count,self.starxcen[i],self.starycen[i])
    #                 # updating the X/Y values after the first iteration
    #                 #  causes problems.  bounces around too much

    #                 if verbose:
    #                     print('Star '+str(i)+' Refitting amp empirically')
    #                     print(str(amp))
                        
    #                 #if i==1: print(amp)
    #                 #if self.niter==2:
    #                 #    import pdb; pdb.set_trace()

    #         # No refit of stellar parameters
    #         else:
    #             model = psf(x,y,pars=[amp,xcen,ycen])

    #         #if self.niter>1:
    #         #    import pdb; pdb.set_trace()
                
    #         # Relculate reduced chi squared
    #         chisq = np.sum((flux-model.ravel())**2/err**2)/npix
    #         self.starchisq[i] = chisq
    #         # chi value, RMS of the residuals as a fraction of the amp
    #         rms = np.sqrt(np.mean(((flux-model.ravel())/self.staramp[i])**2))
    #         self.starrms[i] = rms
            
    #         #model = psf(x,y,pars=[amp,xcen,ycen])
    #         # Zero-out anything beyond the fitting radius
    #         #im[mask] = 0.0
    #         #npix = im.size
    #         #npix = len(x)
    #         allim[pixcnt:pixcnt+npix] = model.flatten()
    #         pixcnt += npix

    #         #import pdb; pdb.set_trace()
            
    #     self.niter += 1
            
    #    return allim


    def fitstars(self,x,psfparams):
        """ Fit amp/xcen/ycen for all the stars using the current PSF."""

        # Loop over the stars and generate the model image
        lookup = np.zeros((1,1,1),float)
        allmodel = np.zeros(self.star_imdata.shape,float)
        maxpix = self.star_imdata.shape[1]
        npix = int(np.sqrt(maxpix))
        for i in range(self.nstars):
            im1,err1,xdata1,ydata1,bbox1,n1 = self.unpackfitstar(i)
            pars = np.zeros(3,float)
            pars[0] = self.staramp[i]
            pars[1] = self.starxcen[i]
            pars[2] = self.starycen[i]
            pout = mnb.psffit(im1,err1,xdata1,ydata1,pars,self.psftype,psfparams,
                              lookup,self.imshape,self.verbose)
            bestpar,perror,cov,flux,fluxerr,chisq = pout
            rchisq = chisq/len(xdata1)
            model,der = mnb.psf(xdata1,ydata1,pars,self.psftype,psfparams,
                                lookup,self.imshape)
            rms = np.sqrt(np.mean(((im1-model)/bestpar[0])**2))
            #print(i,bestpar,rchisq,rms)
            
            # Update stellar parameters
            self.staramp[i] = bestpar[0]
            self.starxcen[i] = bestpar[1]
            self.starycen[i] = bestpar[2]
            self.starchisq[i] = rchisq
            self.starrms[i] = rms
            
    
    def jac(self,x,args):
        """ jacobian."""
        # input the model parameters        
        if self.verbose:
            print('model: '+str(self.niter)+' ',args)
        # Loop over the stars and generate the model image
        nparsarr = np.array([6,7,8,8,7])
        npars = nparsarr[self.psftype-1]
        allmodel = np.zeros(self.star_imdata.shape,float)
        maxpix = self.star_imdata.shape[1]
        allderiv = np.zeros((self.nstars,maxpix,npars),float)
        for i in range(self.nstars):
            im1,err1,xdata1,ydata1,bbox1,n1 = self.unpackfitstar(i)
            allpars = np.zeros(npars,float)
            allpars[0] = self.staramp[i]
            allpars[1] = self.starxcen[i]
            allpars[2] = self.starycen[i]
            allpars[3:] = args
            g,d = mnb.amodel2d(xdata1,ydata1,self.psftype,allpars,npars)
            allmodel[i,:len(g)] = g
            allderiv[i,:len(g),:] = d
        # Trim off the stellar parameter (amp/xc/yc) derivatives
        allderiv = allderiv[:,:,3:]
        return allmodel,allderiv

    def linesearch(self,bestpar,dbeta):
        # Perform line search along search gradient

        # Inside model() the parameters are limited to the PSF bounds()
        f0 = self.chisq(bestpar)
        f1 = self.chisq(bestpar+0.5*dbeta)
        f2 = self.chisq(bestpar+dbeta)
        alpha = quadratic_bisector(np.array([0.0,0.5,1.0]),np.array([f0,f1,f2]))
        alpha = np.minimum(np.maximum(alpha,0.0),1.0)  # 0<alpha<1
        if np.isfinite(alpha)==False:
            alpha = 1.0

        pars_new = bestpar + alpha * dbeta
        new_dbeta = alpha * dbeta
        return alpha,new_dbeta
    
    # def mklookup(self,order=0):
    #     """ Make an empirical look-up table for the residuals."""

    #     # Make the empirical EPSF
    #     cube = self.psf.resid(self.tab,self.image,fillvalue=np.nan)
    #     coords = (self.tab['x'].data,self.tab['y'].data)
    #     epsf,nbadstar,rms = mkempirical(cube,order=order,coords=coords,shape=self.image.shape,lookup=True)
    #     lookup = models.PSFEmpirical(epsf,imshape=self.image.shape,order=order,lookup=True)

    #     # DAOPHOT does some extra analysis to make sure the flux
    #     # in the residual component is okay

    #     # -make sure
    #     #  -to take the total flux into account (not varying across image)
    #     #  -make sure the amp=1 at center
    #     #  -make sure all PSF values are >=0
                             
    #     # Add the lookup table to the PSF model
    #     self.psf.lookup = lookup

    #     #import pdb; pdb.set_trace()
        
        
    def starmodel(self,istar,params):
        """ Generate a 2D star model image that can be compared to the original cutout."""
        im,err,xdata,ydata,bbox,shape = self.unpackstar(istar)
        xdata1d = xdata.ravel()
        ydata1d = ydata.ravel()
        amp = self.staramp[istar]
        xcen = self.starxcen[istar]   
        ycen = self.starycen[istar]
        nparsarr = np.array([6,7,8,8,7])
        npars = nparsarr[self.psftype-1]
        allpars = np.zeros(npars,float)
        allpars[0] = amp
        allpars[1] = xcen
        allpars[2] = ycen
        allpars[3:] = params
        g,_ = mnb.amodel2d(xdata1d,ydata1d,self.psftype,allpars,0)
        g = g.reshape(im.shape)
        return g

@njit
def fitpsf(psftype,psfparams,image,error,starx,stary,starflux,fitradius,method='qr',maxiter=10,
           minpercdiff=1.0,verbose=False):
    """
    Fit PSF model to stars in an image.

    Parameters
    ----------
    psftype : PSF object
       PSF object with initial parameters to use.
    psfparams : numpy array
       PSF object with initial parameters to use.
    image : numpy array
       Image to use to fit PSF model to stars.
    error : numpy array
       Uncertainties in image.
    tab : numpy structured array
       Table with initial amp/x/y values for the stars to use to fit the PSF.
    fitradius : float, table
       The fitting radius.  If none is input then the initial PSF FWHM will be used.
    method : str, optional
       Method to use for solving the non-linear least squares problem: "qr",
       "svd", "cholesky", and "curve_fit".  Default is "qr".
    maxiter : int, optional
       Maximum number of iterations to allow.  Only for methods "qr", "svd", and "cholesky".
       Default is 10.
    minpercdiff : float, optional
       Minimum percent change in the parameters to allow until the solution is
       considered converged and the iteration loop is stopped.  Only for methods
       "qr" and "svd".  Default is 1.0.
    verbose : boolean, optional
       Verbose output.

    Returns
    -------
    pars : numpy array
       Array of best-fit model parameters
    perror : numpy array
       Uncertainties in "pars".
    pcov : numpy array
       Covariances of the best-ft parameters.
    psftab : table
       Table of best-fitting amp/xcen/ycen values for the PSF stars.
    rchisq : float
       Reduced chi-squared of the best-fit.
    pf : PSFFitter object
       PSFFitter object.
    
    Example
    -------

    pars,perror,pcov,psftab,rchisq,pf = fitpsf(psftype,psfparams,image,error,tab,fitradius)

    """

    # Fitting the PSF to the stars
    #-----------------------------

    # Empirical PSF - done differently
    if psf.psftype==6:
        cube1 = starcube(tab,image,npix=psf.npix,fillvalue=np.nan)
    #     coords = (tab['x'].data,tab['y'].data)
    #     epsf1,nbadstar1,rms1 = mkempirical(cube1,order=psf.order,coords=coords,shape=psf._shape)
    #     initpsf = mnb.PSFEmpirical(epsf1,imshape=image.shape,order=psf.order)
    #     pf = PSFFitter(initpsf,image,tab,fitradius=fitradius,verbose=False)
    #     # Fit the amp, xcen, ycen properly
    #     xdata = np.arange(pf.ntotpix)
    #     out = pf.model(xdata,[])
    #     # Put information into the psftab table
    #     psftab['amp'] = pf.staramp
    #     psftab['x'] = pf.starxcen
    #     psftab['y'] = pf.starycen
    #     psftab['chisq'] = pf.starchisq
    #     psftab['rms'] = pf.starrms
    #     psftab['npix'] = pf.starnpix    
    #     for i in range(len(tab)):
    #         bbox = pf.bboxdata[i]
    #         psftab['ixmin'][i] = bbox.ixmin
    #         psftab['ixmax'][i] = bbox.ixmax
    #         psftab['iymin'][i] = bbox.iymin
    #         psftab['iymax'][i] = bbox.iymax        
    #     psftab = Table(psftab)
    #     # Remake the empirical EPSF    
    #     cube = starcube(psftab,image,npix=psf.npix,fillvalue=np.nan)
    #     epsf,nbadstar,rms = mkempirical(cube,order=psf.order,coords=coords,shape=psf._shape)
    #     newpsf = mnb.PSFEmpirical(epsf,imshape=image.shape,order=psf.order)
    #     if verbose:
    #         print('Median RMS: '+str(np.median(pf.starrms)))
    #         print('dt = %.2f sec' % (time.time()-t0))
    #     return newpsf, None, None, psftab, pf

    nstars = len(starx)
    pf = PSFFitter(psftype,psfparams,image,error,starx,stary,starflux,fitradius,verbose)
    initpar = psfparams.copy()
    xdata = np.arange(10)  # dummy

    # Iterate
    niter = 0
    percdiff = 1e10
    bestpar = initpar.copy()
    
    dchisq = -1
    oldchisq = 1e30
    bounds = mnb.model2d_bounds(psftype)
    bounds = bounds[3:,:]
    while (niter<maxiter and percdiff>minpercdiff and dchisq<0):
        # Fit the stellar parameters first
        pf.fitstars(xdata,bestpar)
        # Get the Jacobian and model
        m,jac = pf.jac(xdata,bestpar)
        chisq = np.sum((pf.starfit_imdata-m)**2/pf.starfit_errdata**2)
        dy = (pf.starfit_imdata-m).ravel()
        # Weights
        wt = 1/pf.starfit_errdata**2
        wt = wt.ravel()
        jac = jac.copy().reshape((jac.shape[0]*jac.shape[1],jac.shape[2]))
        m = m.ravel()
        # Solve Jacobian
        dbeta = mnb.qr_jac_solve(jac,dy,weight=wt)

        # Perform line search
        alpha,new_dbeta = pf.linesearch(bestpar,dbeta)
            
        if verbose:
            print('  pars = '+str(bestpar))
            print('  dbeta = '+str(dbeta))

        # Update parameters
        oldpar = bestpar.copy()
        # limit the steps to the maximum step sizes and boundaries
        maxsteps = mnb.model2d_maxsteps(psftype,bestpar)
        maxsteps = maxsteps[3:]
        bestpar = mnb.newlsqpars(bestpar,new_dbeta,bounds,maxsteps)
        
        # Update the parameters 
        diff = np.abs(bestpar-oldpar)
        denom = np.abs(oldpar.copy())
        bad, = np.where(denom==0)
        if len(bad)>0:
            denom[bad] = 1.0  # deal with zeros
        percdiff = np.max(diff/denom*100)
        dchisq = chisq-oldchisq
        percdiffchisq = dchisq/oldchisq*100
        oldchisq = chisq
        niter += 1
            
        if verbose:
            print('  '+str(niter+1)+' ',bestpar,' '+str(percdiff)+' '+str(chisq))

    # Fit the stellar parameters first
    pf.fitstars(xdata,bestpar)
    # Get the Jacobian and model
    m,jac = pf.jac(xdata,bestpar)
    chisq = np.sum((pf.starfit_imdata-m)**2/pf.starfit_errdata**2)
    dy = (pf.starfit_imdata-m).ravel()
    # Weights
    wt = 1/pf.starfit_errdata**2
    wt = wt.ravel()
    jac = jac.copy().reshape((jac.shape[0]*jac.shape[1],jac.shape[2]))
    rchisq = chisq/np.sum(pf.starfit_ndata)
    
    # Estimate uncertainties
    # Calculate covariance matrix
    pcov = mnb.jac_covariance(jac,dy,wt=wt)
    perror = np.sqrt(np.diag(pcov))

    pars = bestpar
    if verbose:
        print('Best-fitting parameters: '+str(pars))
        print('Errors: '+str(perror))
        print('Median RMS: '+str(np.median(pf.starrms)))

    # Output best-fitting values for the PSF stars as well
    psftab = np.zeros((nstars,11),float)
    psftab[:,0] = np.arange(nstars)+1.0  # id
    psftab[:,1] = pf.staramp             # amp
    psftab[:,2] = pf.starxcen            # x
    psftab[:,3] = pf.starycen            # y
    psftab[:,4] = pf.starfit_ndata       # npix
    psftab[:,5] = pf.starrms             # rms
    psftab[:,6] = pf.starchisq           # chisq
    psftab[:,7] = pf.star_bbox[:,0]      # ixmin
    psftab[:,8] = pf.star_bbox[:,1]      # ixmax
    psftab[:,9] = pf.star_bbox[:,2]      # iymin
    psftab[:,10] = pf.star_bbox[:,3]     # iymax
    
    return pars, perror, pcov, psftab, rchisq
    

#@njit
def getpsf(psftype,psfparams,image,error,starx,stary,starflux,fitradius,
           lookup=False,lorder=0,subnei=False,
           alltab=None,maxiter=10,minpercdiff=1.0,reject=False,maxrejiter=3,verbose=False):
    """
    Fit PSF model to stars in an image with outlier rejection of badly-fit stars.

    Parameters
    ----------
    psftype : PSF object
       PSF object with initial parameters to use.
    psfparams : numpy array
       PSF object with initial parameters to use.
    image : numpy array
       Image to use to fit PSF model to stars.
    error : numpy array
       Uncertainties in image.
    tab : table
       Catalog with initial amp/x/y values for the stars to use to fit the PSF.
    fitradius : float, table
       The fitting radius.  If none is input then the initial PSF FWHM will be used.
    lookup : boolean, optional
       Use an empirical lookup table.  Default is False.
    lorder : int, optional
       The order of the spatial variations (0=constant, 1=linear).  Default is 0.
    method : str, optional
       Method to use for solving the non-linear least squares problem: "qr",
       "svd", "cholesky", and "curve_fit".  Default is "qr".
    subnei : boolean, optional
       Subtract stars neighboring the PSF stars.  Default is False.
    alltab : table, optional
       Catalog of all objects in the image.  This is needed for bad PSF star
       rejection.
    maxiter : int, optional
       Maximum number of iterations to allow.  Only for methods "qr", "svd", and "cholesky".
       Default is 10.
    minpercdiff : float, optional
       Minimum percent change in the parameters to allow until the solution is
       considered converged and the iteration loop is stopped.  Only for methods
       "qr" and "svd".  Default is 1.0.
    reject : boolean, optional
       Reject PSF stars with high RMS values.  Default is False.
    maxrejiter : int, boolean
       Maximum number of PSF star rejection iterations.  Default is 3.
    verbose : boolean, optional
       Verbose output.

    Returns
    -------
    newpsf : PSF object
       New PSF object with the best-fit model parameters.
    pars : numpy array
       Array of best-fit model parameters
    perror : numpy array
       Uncertainties in "pars".
    psftab : table
       Table of best-fitting amp/xcen/ycen values for the PSF stars.

    Example
    -------

    newpsf,pars,perror,psftab = getpsf(psf,image,tab)

    """

    #t0 = time.time()
    #print = utils.getprintfunc() # Get print function to be used locally, allows for easy logging   

    #psftype,psfpars,_,_ = mnb.unpackpsf(psf)
    
    # # Fitting radius
    # if fitradius is None:
    #     tpars = np.zeros(len(psfpars)+3,float)
    #     tpars[0] = 1.0
    #     tpars[3:] = psfpars
    #     if psftype == 3:  # Penny
    #         fitradius = mnb.penny2d_fwhm(tpars)*1.5
    #     else:
    #         fitradius = mnb.model2d_fwhm(psftype,tpars)
        
    # # subnei but no alltab input
    # if subnei and alltab is None:
    #     raise ValueError('alltab is needed for PSF neighbor star subtraction')
        
    # if 'id' not in tab.dtype.names:
    #     tab['id'] = np.arange(len(tab))+1
    # psftab = tab.copy()

    # id, amp, x, y, npix, rms, chisq, ixmin, ixmax, iymin, iymax, reject
    psftab = np.zeros((nstars,12),float)
    psftab[:,0] = np.arange(nstars)+1
    psftab[:,1] = starflux
    psftab[:,2] = starx
    psftab[:,3] = stary
    
    # Initializing output PSF star catalog
    #dt = np.dtype([('id',int),('amp',float),('x',float),('y',float),('npix',int),
    #               ('rms',float),('chisq',float),('ixmin',int),('ixmax',int),
    #               ('iymin',int),('iymax',int),('reject',int)])
    #outtab = np.zeros(len(tab),dtype=dt)
    #outtab = Table(outtab)
    #for n in ['id','x','y']:
    #    outtab[n] = tab[n]
    nstars = len(starx)
    # id, amp, x, y, npix, rms, chisq, ixmin, ixmax, iymin, iymax, reject
    outtab = np.zeros((nstars,12),float)
    
    # Remove stars that are too close to the edge
    ny,nx = image.shape
    gd, = np.where((starx>fitradius) & (starx<(nx-1-fitradius)) &
                   (stary>fitradius) & (stary<(ny-1-fitradius)))
    ngd = len(gd)
    nbd = nstars-ngd
    if nbd > 0:
        if verbose:
            print('Removing '+str(nbd)+' stars near the edge')
        psftab = psftab[gd,:]

    # # Generate an empirical image of the stars
    # # and fit a model to it to get initial estimates
    # if psftype != 6:
    #     cube = starcube(psftab,image,npix=psf.npix,fillvalue=np.nan)
    #     epsf,nbadstar,rms = mkempirical(cube,order=0)
    #     #epsfim = CCDData(epsf,error=epsf.copy()*0+1,mask=~np.isfinite(epsf))
    #     epsfim = epsf.copy()
    #     epsferr = np.ones(epsf.shape,float)
    #     ny,nx = epsf.shape
    #     xx,yy = np.meshgrid(np.arange(nx),np.arange(ny))
    #     out = model2dfit(epsfim,epsferr,xx,yy,psftype,1.0,nx//2,ny//2,verbose=False)
    #     pars,perror,cov,flux,fluxerr,chisq = out
    #     mparams = pars[3:]  # model parameters
    #     #pars,perror,mparams = mnb.model2d_fit(epsfim,pars=[1.0,psf.npix/2,psf.npix//2])
    #     initpar = mparams.copy()
    #     curpsf = mnb.packpsf(psftype,mparams,0,0)
    #     #curpsf = psf.copy()
    #     #curpsf.params = initpar
    #     if verbose:
    #         print('Initial estimate from empirical PSF fit = '+str(mparams))
    # else:
    #     curpsf = psf.copy()
    #     _,initpar,_,_ = mnb.unpackpsf(psf)
    #     #initpar = psf.params.copy()
    curparams = psfparams.copy()
    
    # Outlier rejection iterations
    nrejiter = 0
    flag = 0
    nrejstar = 100
    fitrad = fitradius
    useimage = image.copy()
    while (flag==0):
        if verbose:
            print('--- Iteration '+str(nrejiter+1)+' ---')                

        # Update the fitting radius
        #if nrejiter>0:
        #    fitrad = curpsf.fwhm()
        if verbose:
            print('  Fitting radius = %5.3f' % (fitrad))
                    
        # Reject outliers
        if reject and nrejiter>0:
            medrms = np.median(ptab[:,5]) # rms
            sigrms = utils.mad(ptab[:,5])
            gd, = np.where(ptab[:,5] < medrms+3*sigrms)
            nrejstar = len(psftab)-len(gd)
            if verbose:
                print('  RMS = %6.4f +/- %6.4f' % (medrms,sigrms))
                print('  Threshold RMS = '+str(medrms+3*sigrms))
                print('  Rejecting '+str(nrejstar)+' stars')
            if nrejstar>0:
                psftab = psftab[gd,:]

        # Subtract neighbors
        if nrejiter>0 and subnei:
            if verbose:
                print('Subtracting neighbors')
                # Find the neighbors in alltab
                # Fit the neighbors and PSF stars
                # Subtract neighbors from the image
                useimage = image.copy()  # start with original image
                print('subtractnei')
                #useimage = subtractnei(useimage,alltab,tab,curpsf)
                
        # Fitting the PSF to the stars
        #-----------------------------
        newpars,perror,pcov,ptab,rchisq = fitpsf(psftype,curparams,image,error,psftab[:,2],psftab[:,3],
                                                 psftab[:,1],fitradius,method=method,
                                                 maxiter=maxiter,minpercdiff=minpercdiff,verbose=verbose)
        #newpsf,pars,perror,ptab,pf = fitpsf(curpsf,useimage,psftab,fitradius=fitrad,method=method,
        #                                    maxiter=maxiter,minpercdiff=minpercdiff,verbose=verbose)
        
        # Add information into the output catalog
        _,ind1,ind2 = np.intersect1d(outtab[:,0],ptab[:,0],return_indices=True)
        #ind1,ind2 = dln.match(outtab['id'],ptab['id'])
        outtab[:,11] = 1    # reject
        #outtab['reject'] = 1
        #for n in ptab.columns:
        #    outtab[n][ind1] = ptab[n][ind2]
        #outtab['reject'][ind1] = 0
        # id, amp, x, y, npix, rms, chisq, ixmin, ixmax, iymin, iymax, reject
        outtab[ind1,1:10] = ptab[ind2,1:10]
        
        # Compare PSF parameters
        # if type(newpsf)!=mnb.PSFEmpirical:
        #     pardiff = newpsf.params-curpsf.params
        # else:
        #     pardiff = newpsf._data-curpsf._data
        pardiff = newpars - pars
        sumpardiff = np.sum(np.abs(pardiff))
        curpsf = newpsf.copy()
        
        # Stopping criteria
        if reject is False or sumpardiff<0.05 or nrejiter>=maxrejiter or nrejstar==0: flag=1
        if subnei is True and nrejiter==0: flag=0   # iterate at least once with neighbor subtraction
        
        nrejiter += 1
        
    # # Generate an empirical look-up table of corrections
    # if lookup:
    #     if verbose:
    #         print('Making empirical lookup table with order='+str(lorder))

    #     pf.mklookup(lorder)
    #     # Fit the stars again and get new RMS values
    #     xdata = np.arange(pf.ntotpix)
    #     out = pf.model(xdata,*pf.psf.params)
    #     newpsf = pf.psf.copy()
    #     # Update information in the output catalog
    #     ind1,ind2 = dln.match(outtab['id'],ptab['id'])
    #     outtab['reject'] = 1
    #     outtab['reject'][ind1] = 0
    #     outtab['amp'][ind1] = pf.staramp[ind2]
    #     outtab['x'][ind1] = pf.starxcen[ind2]
    #     outtab['y'][ind1] = pf.starycen[ind2]
    #     outtab['rms'][ind1] = pf.starrms[ind2]
    #     outtab['chisq'][ind1] = pf.starchisq[ind2]                
    #     if verbose:
    #         print('Median RMS: '+str(np.median(pf.starrms)))    
            
    
    return newpsf, pars, perror, outtab
        
        



# import os
import numpy as np
# import numba
from numba import njit,types,from_dtype,typed
# from numba.typed import Dict,List
from numba.experimental import jitclass
from numba_kdtree import KDTree

# Fit a PSF model to multiple stars in an image

PI = 3.141592653589793

from numba.pycc import CC
cc = CC('coords_numba_static')

            
# # from astroML
@njit
@cc.export('crossmatch', '(f8[:,:],f8[:,:],f8,i8)')
def crossmatch(X1, X2, max_distance=np.inf,k=1):
    """Cross-match the values between X1 and X2

    By default, this uses a KD Tree for speed.

    Parameters
    ----------
    X1 : array_like
        first dataset, shape(N1, D)
    X2 : array_like
        second dataset, shape(N2, D)
    max_distance : float (optional)
        maximum radius of search.  If no point is within the given radius,
        then inf will be returned.

    Returns
    -------
    dist, ind: ndarrays
        The distance and index of the closest point in X2 to each point in X1
        Both arrays are length N1.
        Locations with no match are indicated by
        dist[i] = inf, ind[i] = N2
    """
    #X1 = np.asarray(X1, dtype=float)
    #X2 = np.asarray(X2, dtype=float)

    N1, D1 = X1.shape
    N2, D2 = X2.shape

    if D1 != D2:
        raise ValueError('Arrays must have the same second dimension')

    kdt = KDTree(X2)

    dist, ind, neigh = kdt.query(X1, k=k, distance_upper_bound=max_distance)

    return dist, ind
    

# from astroML, modified by D. Nidever
@njit
@cc.export('xmatch', '(f8[:],f8[:],f8[:],f8[:],f8,b1,b1)')
def xmatch(ra1, dec1, ra2, dec2, dcr=2.0, unique=False, sphere=True):
    """Cross-match angular values between RA1/DEC1 and RA2/DEC2

    Find the closest match in the second list for each element
    in the first list and within the maximum distance.

    By default, this uses a KD Tree for speed.  Because the
    KD Tree only handles cartesian distances, the angles
    are projected onto a 3D sphere.

    This can return duplicate matches if there is an element
    in the second list that is the closest match to two elements
    of the first list.

    Parameters
    ----------
    ra1/dec1 : array_like
        first dataset, arrays of RA and DEC
        both measured in degrees
    ra2/dec2 : array_like
        second dataset, arrays of RA and DEC
        both measured in degrees
    dcr : float (optional)
        maximum radius of search, measured in arcsec.
        This can be an array of the same size as ra1/dec1.
    unique : boolean, optional
        Return unique one-to-one matches.  Default is False and
           allows duplicates.
    sphere : boolean, optional
        The coordinates are spherical in degrees.  Otherwise, the dcr
          is assumed to be in the same units as the input values.
          Default is True.


    Returns
    -------
    ind1, ind2, dist: ndarrays
        The indices for RA1/DEC1 (ind1) and for RA2/DEC2 (ind2) of the
        matches, and the distances (in arcsec).
    """
    n1 = len(ra1)
    n2 = len(ra2)
    X1 = np.zeros((n1,2),float)
    X1[:,0] = ra1
    X1[:,1] = dec1
    X2 = np.zeros((n2,2),float)
    X2[:,0] = ra2
    X2[:,1] = dec2
    
    # Spherical coordinates in degrees
    if sphere:
        X1 = X1 * (np.pi / 180.)
        X2 = X2 * (np.pi / 180.)
        #if utils.size(dcr)>1:
        #    max_distance = (np.max(dcr) / 3600) * (np.pi / 180.)
        #else:
        #    max_distance = (dcr / 3600) * (np.pi / 180.)
        max_distance = (dcr / 3600) * (np.pi / 180.)
        
        # Convert 2D RA/DEC to 3D cartesian coordinates
        Y1 = np.zeros((n1,3),float)
        Y1[:,0] = np.cos(X1[:, 0]) * np.cos(X1[:, 1])
        Y1[:,1] = np.sin(X1[:, 0]) * np.cos(X1[:, 1])
        Y1[:,2] = np.sin(X1[:, 1])
        #Y1 = np.transpose(np.vstack([np.cos(X1[:, 0]) * np.cos(X1[:, 1]),
        #                             np.sin(X1[:, 0]) * np.cos(X1[:, 1]),
        #                             np.sin(X1[:, 1])]))
        Y2 = np.zeros((n2,3),float)
        Y2[:,0] = np.cos(X2[:, 0]) * np.cos(X2[:, 1])
        Y2[:,1] = np.sin(X2[:, 0]) * np.cos(X2[:, 1])
        Y2[:,2] = np.sin(X2[:, 1])
        #Y2 = np.transpose(np.vstack([np.cos(X2[:, 0]) * np.cos(X2[:, 1]),
        #                             np.sin(X2[:, 0]) * np.cos(X2[:, 1]),
        #                             np.sin(X2[:, 1])]))

        # law of cosines to compute 3D distance
        max_y = np.sqrt(2 - 2 * np.cos(max_distance))
        k = 1 if unique is False else 10
        dist, ind = crossmatch(Y1, Y2, max_y, k=k)
        # dist has shape [N1,10] or [N1,1] (if unique)
    
        # convert distances back to angles using the law of tangents
        not_infy,not_infx = np.where(~np.isinf(dist))
        #x = 0.5 * dist[not_infy,not_infx]
        #dist[not_infy,not_infx] = (180. / np.pi * 2 * np.arctan2(x,
        #                           np.sqrt(np.maximum(0, 1 - x ** 2))))
        #dist[not_infy,not_infx] *= 3600.0      # in arcsec
    # Regular coordinates
    else:
        k = 1 if unique is False else 10
        dist, ind = crossmatch(X1, X2, dcr, k=k)
        #dist, ind = crossmatch(X1, X2, np.max(dcr), k=k)
        not_infy,not_infx = np.where(~np.isinf(dist))
        
    # Allow duplicates
    if unique==False:

        # no matches
        if len(not_infx)==0:
            return np.array([-1]), np.array([-1]), np.array([np.inf])
        
        # If DCR is an array then impose the max limits for each element
        #if utils.size(dcr)>1:
        #    bd,nbd = utils.where(dist > dcr)
        #    if nbd>0:
        #        dist[bd] = np.inf
        #        not_inf = ~np.isinf(dist)
        
        # Change to the output that I want
        # dist is [N1,1] if unique==False
        ind1 = np.arange(len(ra1))[not_infy]
        ind2 = ind[not_infy,0]
        mindist = dist[not_infy,0]
        
    # Return unique one-to-one matches
    else:

        # no matches
        if np.sum(~np.isinf(dist[:,0]))==0:
            return np.array([-1]), np.array([-1]), np.array([np.inf])
        
        done = 0
        niter = 1
        # Loop until we converge
        while (done==0):

            # If DCR is an array then impose the max limits for each element
            #if utils.size(dcr)>1:
            #    bd,nbd = utils.where(dist[:,0] > dcr)
            #    if nbd>0:
            #        for i in range(nbd):
            #            dist[bd[i],:] = np.inf

            # no matches
            if np.sum(~np.isinf(dist[:,0]))==0:
                return np.array([-1]), np.array([-1]), np.array([np.inf])

            # closest matches
            not_inf1 = ~np.isinf(dist[:,0])
            not_inf1_ind, = np.where(~np.isinf(dist[:,0]))
            ind1 = np.arange(len(ra1))[not_inf1]  # index into original ra1/dec1 arrays
            ind2 = ind[:,0][not_inf1]             # index into original ra2/dec2 arrays
            mindist = dist[:,0][not_inf1]
            if len(ind2)==0:
                return np.array([-1]), np.array([-1]), np.array([np.inf])
            find2 = np.zeros(len(ind2),float)
            find2[:] = ind2
            index = Index(find2)
            # some duplicates to deal with
            bd, = np.where(index.num>1)
            nbd = len(bd)
            if nbd>0:
                ntorem = 0
                for i in range(nbd):
                    ntorem += index.num[bd[i]]-1
                torem = np.zeros(ntorem,np.int32)  # index into shortened ind1/ind2/mindist
                trcount = 0
                for i in range(nbd):
                    # index into shortened ind1/ind2/mindist
                    indx = index.getindex(bd[i])
                    #indx = index['index'][index['lo'][bd[i]]:index['hi'][bd[i]]+1]
                    # keep the one with the smallest minimum distance
                    si = np.argsort(mindist[indx])
                    if index.num[bd[i]]>2:
                        bad = indx[si[1:]]
                        torem[trcount:trcount+len(bad)] = bad    # add
                        trcount += len(bad)
                    else:
                        torem[trcount:trcount+1] = indx[si[1:]][0]  # add single element
                #ntorem = utils.size(torem)
                torem_orig_index = not_inf1_ind[torem]  # index into original ind/dist arrays
                # For each object that was "removed" and is now unmatched, check the next possible
                # match and move it up in the dist/ind list if it isn't INF
                for i in range(ntorem):
                    # There is a next possible match 
                    if ~np.isinf(dist[torem_orig_index[i],niter-1]):
                        temp = np.zeros(10,np.int64)
                        temp[niter:] = ind[torem_orig_index[i],niter:]  #.squeeze()
                        temp[-niter:] = np.zeros(niter,np.int64)-1
                        ind[torem_orig_index[i],:] = temp
                        temp2 = np.zeros(10,float)
                        temp2[niter:] = dist[torem_orig_index[i],niter:]   #.squeeze()
                        temp2[-niter:] = np.zeros(niter,float)+np.inf
                        dist[torem_orig_index[i],:] = temp2
                        #ind[torem_orig_index[i],:] = np.hstack( (ind[torem_orig_index[i],niter:].squeeze(),
                        #                                         np.repeat(-1,niter)) )
                        #dist[torem_orig_index[i],:] = np.hstack( (dist[torem_orig_index[i],niter:].squeeze(),
                        #                                          np.repeat(np.inf,niter)) )
                    # All INFs
                    else:
                        ind[torem_orig_index[i],:] = -1
                        dist[torem_orig_index[i],:] = np.inf
                        # in the next iteration these will be ignored
            else:
                ntorem = 0

            niter += 1
            # Are we done, no duplicates or hit the maximum 10
            if (ntorem==0) or (niter>=10): done=1
                                
    return ind1, ind2, mindist
